"""This module contains the base class of all music providers."""

from __future__ import annotations

from typing import Optional

from core.settings import storage
from core.celery import app
from core.models import ArchivedSong
from core.musiq import musiq, playback


class ProviderError(Exception):
    """An error to indicate that an error occurred while providing music."""


class WrongUrlError(Exception):
    """An error to indicate that a provider was called
    with a url that belongs to a different service."""


class MusicProvider:
    """The base class for all music providers.
    Provides abstract function declarations."""

    def __init__(self, query: Optional[str], key: Optional[int]) -> None:
        self.query = query
        self.key = key
        if not hasattr(self, "type"):
            # the type should already have been set by the base class
            self.type = "unknown"
            assert False
        self.id: Optional[str] = self.extract_id()
        self.ok_message = "ok"
        self.error = "error"

    def extract_id(self) -> Optional[str]:
        """Tries to extract the id from the given query.
        Returns the id if possible, otherwise None"""
        return None

    def check_cached(self) -> bool:
        """Returns whether this resource is available on disk."""
        raise NotImplementedError()

    def check_available(self) -> bool:
        """Returns whether this resource is available online."""
        raise NotImplementedError()

    def enqueue_placeholder(self, manually_requested) -> None:
        """Enqueues a placeholder if applicable. Playlists have no placeholder, only songs do.
        Used to identify this resource in the client after a request."""
        raise NotImplementedError()

    def remove_placeholder(self) -> None:
        """Removes the placeholder in the queue that represents this resource.
        Called if there was an error and this element needs to be removed from the queue."""
        raise NotImplementedError()

    def make_available(self) -> bool:
        """Makes this resource available for playback.
        If possible, downloads it to disk.
        If this takes a long time, calls update_state so the placeholder is visible.
        Returns False if an error occured, True otherwise."""
        raise NotImplementedError()

    def persist(self, request_ip: str, archive: bool = True) -> None:
        """Updates the database.
        Creates an archived entry or updates it.
        Also handles logging to database."""
        raise NotImplementedError()

    def enqueue(self) -> None:
        """Updates the placeholder in the song queue with the actual data."""
        raise NotImplementedError()

    def request(
        self, request_ip: str, archive: bool = True, manually_requested: bool = True
    ) -> None:
        """Tries to request this resource.
        Uses the local cache if possible, otherwise tries to retrieve it online."""

        if 0 < storage.get("max_queue_length") <= playback.queue.count():
            self.error = "Queue limit reached"
            raise ProviderError(self.error)

        enqueue_function = enqueue

        if not self.check_cached():
            if self.query is not None and storage.get("additional_keywords"):
                # add the additional keywords from the settings before checking
                self.query += " " + storage.get("additional_keywords")
            if not self.check_available():
                raise ProviderError(self.error)

            # overwrite the enqueue function and make the resource available before calling it
            enqueue_function = fetch_enqueue

        from core.musiq.song_provider import SongProvider

        if storage.get("new_music_only") and isinstance(self, SongProvider):
            try:
                archived_song = ArchivedSong.objects.get(url=self.get_external_url())
                if archived_song.counter > 0:
                    self.error = "Only new music is allowed!"
                    raise ProviderError(self.error)
            except ArchivedSong.DoesNotExist:
                pass

        self.enqueue_placeholder(manually_requested)

        enqueue_function.delay(self, request_ip, archive)


@app.task
def enqueue(provider: MusicProvider, request_ip: str, archive: bool) -> None:
    """Enqueue the music managed by the given provider."""
    provider.persist(request_ip, archive=archive)
    provider.enqueue()


@app.task
def fetch_enqueue(provider: MusicProvider, request_ip: str, archive: bool) -> None:
    """Fetch and enqueue the music managed by the given provider."""
    if not provider.make_available():
        provider.remove_placeholder()
        musiq.update_state()
        return

    enqueue(provider, request_ip, archive)



# Generated by Django 2.2.3 on 2019-09-12 20:19

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [("core", "0002_auto_20190909_1707")]

    operations = [
        migrations.CreateModel(
            name="ArchivedPlaylist",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("list_id", models.CharField(max_length=200, unique=True)),
                ("title", models.CharField(max_length=1000)),
                ("counter", models.IntegerField()),
            ],
        ),
        migrations.CreateModel(
            name="PlaylistEntry",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("index", models.IntegerField()),
                ("url", models.CharField(max_length=200)),
                (
                    "playlist",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="entries",
                        to="core.ArchivedPlaylist",
                    ),
                ),
                (
                    "song",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="playlist_entries",
                        to="core.ArchivedSong",
                    ),
                ),
            ],
            options={"ordering": ["playlist", "index"]},
        ),
        migrations.CreateModel(
            name="ArchivedPlaylistQuery",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("query", models.CharField(max_length=1000)),
                (
                    "playlist",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="queries",
                        to="core.ArchivedPlaylist",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="requestlog",
            name="playlist",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="core.ArchivedPlaylist",
            ),
        ),
    ]



"""This module analyses provides an analysis of gathered data."""
from __future__ import annotations

import math
from datetime import datetime, timedelta
from typing import Tuple

from dateutil import tz
from django.conf import settings as conf
from django.core.handlers.wsgi import WSGIRequest
from django.db import models
from django.http import HttpResponse
from django.http import HttpResponseBadRequest
from django.http import JsonResponse
from django.utils import dateparse
from django.utils import timezone

import core.musiq.song_utils as song_utils
from core.models import PlayLog, ArchivedPlaylist, PlaylistEntry
from core.models import RequestLog
from core.settings.settings import control


def _parse_datetimes(request: WSGIRequest) -> Tuple[datetime, datetime]:
    if request.method == "GET":
        params = request.GET
    else:
        params = request.POST

    startdate = params.get("startdate")
    starttime = params.get("starttime")
    enddate = params.get("enddate")
    endtime = params.get("endtime")
    if not startdate or not starttime or not enddate or not endtime:
        raise ValueError("All fields are required")

    start = dateparse.parse_datetime(startdate + "T" + starttime)
    end = dateparse.parse_datetime(enddate + "T" + endtime)

    if start is None or end is None:
        raise ValueError("invalid start-/endtime given")
    if start >= end:
        raise ValueError("start has to be before end")

    start = timezone.make_aware(start)
    end = timezone.make_aware(end)

    return start, end


@control
def analyse(request: WSGIRequest) -> HttpResponse:
    """Perform an analysis of the database in the given timeframe."""

    try:
        start, end = _parse_datetimes(request)
    except ValueError as e:
        return HttpResponseBadRequest(e.args[0])

    played = PlayLog.objects.all().filter(created__gte=start).filter(created__lt=end)
    if not played.exists():
        return HttpResponseBadRequest("No songs played in the given time span")

    requested = (
        RequestLog.objects.all().filter(created__gte=start).filter(created__lt=end)
    )
    played_count = (
        played.values("song__url", "song__artist", "song__title")
        .values(
            "song__url", "song__artist", "song__title", count=models.Count("song__url")
        )
        .order_by("-count")
    )
    played_votes = (
        PlayLog.objects.all()
        .filter(created__gte=start)
        .filter(created__lt=end)
        .order_by("-votes")
    )
    devices = requested.values("address").values(
        "address", count=models.Count("address")
    )

    response = {
        "songsPlayed": len(played),
        "mostPlayedSong": (
            song_utils.displayname(
                played_count[0]["song__artist"], played_count[0]["song__title"]
            )
            + f" ({played_count[0]['count']})"
        ),
        "highestVotedSong": (
            played_votes[0].song_displayname() + f" ({played_votes[0].votes})"
        ),
        "mostActiveDevice": (devices[0]["address"] + f" ({devices[0]['count']})"),
    }
    requested_by_ip = requested.filter(address=devices[0]["address"])
    for i in range(6):
        if i >= len(requested_by_ip):
            break
        response["mostActiveDevice"] += "\n"
        if i == 5:
            response["mostActiveDevice"] += "..."
        else:
            response["mostActiveDevice"] += requested_by_ip[i].item_displayname()

    binsize = 3600
    number_of_bins = math.ceil((end - start).total_seconds() / binsize)
    request_bins = [0 for _ in range(number_of_bins)]

    for request_log in requested:
        seconds = (request_log.created - start).total_seconds()
        index = int(seconds / binsize)
        request_bins[index] += 1

    current_time = start
    current_index = 0
    response["requestActivity"] = ""
    while current_time < end:
        response["requestActivity"] += current_time.strftime("%H:%M")
        response["requestActivity"] += ":\t" + str(request_bins[current_index])
        response["requestActivity"] += "\n"
        current_time += timedelta(seconds=binsize)
        current_index += 1

    localtz = tz.gettz(conf.TIME_ZONE)
    playlist = ""
    for play_log in played:
        localtime = play_log.created.astimezone(localtz)
        playlist += "[{:02d}:{:02d}] {}\n".format(
            localtime.hour, localtime.minute, play_log.song_displayname()
        )
    response["playlist"] = playlist

    return JsonResponse(response)


@control
def save_as_playlist(request: WSGIRequest) -> HttpResponse:
    """Save the songs in the given timeframe as a playlist with the given name."""

    try:
        start, end = _parse_datetimes(request)
    except ValueError as e:
        return HttpResponseBadRequest(e.args[0])

    name = request.POST.get("name")
    if not name:
        return HttpResponseBadRequest("Name required")

    played = PlayLog.objects.all().filter(created__gte=start).filter(created__lt=end)

    list_id = f"playlog {str(start).replace(' ','T')} {str(end).replace(' ', 'T')}"

    playlist, created = ArchivedPlaylist.objects.get_or_create(
        list_id=list_id, title=name, counter=0
    )
    if not created:
        return HttpResponseBadRequest("Playlist already exists")

    song_index = 0
    for log in played:
        external_url = log.song.url
        PlaylistEntry.objects.create(
            playlist=playlist, index=song_index, url=external_url
        )
        song_index += 1

    return HttpResponse()



# Generated by Django 2.2.3 on 2019-09-09 15:07

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("core", "0001_initial")]

    operations = [
        migrations.AddField(
            model_name="currentsong",
            name="manually_requested",
            field=models.BooleanField(default=True),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="playlog",
            name="manually_requested",
            field=models.BooleanField(default=True),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="queuedsong",
            name="manually_requested",
            field=models.BooleanField(default=True),
            preserve_default=False,
        ),
    ]



"""This module manages and counts user accesses and handles permissions."""
import time
from typing import Any, Callable

import ipware
from django.contrib.auth.models import AbstractUser
from django.core.handlers.wsgi import WSGIRequest

import core.settings.storage as storage
from core import redis

# kick users after some time without any request
INACTIVITY_PERIOD = 600


def has_controls(user: AbstractUser) -> bool:
    """Determines whether the given user is allowed to control playback."""
    return user.username == "mod" or is_admin(user)


def is_admin(user: AbstractUser) -> bool:
    """Determines whether the given user is the admin."""
    return user.is_superuser


def update_user_count() -> None:
    """Go through all recent requests and delete those that were too long ago."""
    now = time.time()
    last_requests = redis.get("last_requests")
    for key, value in list(last_requests.items()):
        if now - value >= INACTIVITY_PERIOD:
            del last_requests[key]
            redis.set("last_requests", last_requests)
    redis.set("last_user_count_update", now)


def get_count() -> int:
    """Returns the number of currently active users.
    Updates this number after an intervals since the last update."""
    if time.time() - redis.get("last_user_count_update") >= 60:
        update_user_count()
    return len(redis.get("last_requests"))


def partymode_enabled() -> bool:
    """Determines whether partymode is enabled,
    based on the number of currently active users."""
    return len(redis.get("last_requests")) >= storage.get("people_to_party")


def get_client_ip(request: WSGIRequest):
    if not storage.get("logging_enabled"):
        return ""
    request_ip, _ = ipware.get_client_ip(request)
    if request_ip is None:
        request_ip = ""
    return request_ip


class SimpleMiddleware:
    """This middleware tracks stores the last access for every connected ip
    so the number of active users can be determined."""

    def __init__(self, get_response: Callable[[WSGIRequest], Any]) -> None:
        # One-time configuration and initialization.
        self.get_response = get_response

    def __call__(self, request: WSGIRequest) -> Any:
        # Code to be executed for each request before
        # the view (and later middleware) are called.
        request_ip = get_client_ip(request)
        last_requests = redis.get("last_requests")
        last_requests[request_ip] = time.time()
        redis.set("last_requests", last_requests)

        response = self.get_response(request)

        # Code to be executed for each request/response after
        # the view is called.

        return response



"""This module handles the basic endpoints led and screen visualizations."""

from __future__ import annotations

from typing import Dict, Any

from django.contrib.auth.decorators import user_passes_test
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse
from django.shortcuts import render

from core import user_manager, base, redis
from core.settings import storage
from core.state_handler import send_state


def state_dict() -> Dict[str, Any]:
    state = base.state_dict()

    lights_state = {}
    lights_state["ringConnected"] = redis.get("ring_initialized")
    lights_state["ringProgram"] = storage.get("ring_program")
    lights_state["ringBrightness"] = storage.get("ring_brightness")
    lights_state["ringMonochrome"] = storage.get("ring_monochrome")
    lights_state["wledLedCount"] = storage.get("wled_led_count")
    lights_state["wledIp"] = storage.get("wled_ip")
    lights_state["wledPort"] = storage.get("wled_port")
    lights_state["wledConnected"] = redis.get("wled_initialized")
    lights_state["wledProgram"] = storage.get("wled_program")
    lights_state["wledBrightness"] = storage.get("wled_brightness")
    lights_state["wledMonochrome"] = storage.get("wled_monochrome")
    lights_state["stripConnected"] = redis.get("strip_initialized")
    lights_state["stripProgram"] = storage.get("strip_program")
    lights_state["stripBrightness"] = storage.get("strip_brightness")
    lights_state["screenConnected"] = redis.get("screen_initialized")
    lights_state["screenProgram"] = storage.get("screen_program")
    lights_state["programSpeed"] = storage.get("program_speed")
    lights_state["fixedColor"] = "#{:02x}{:02x}{:02x}".format(
        *(int(val * 255) for val in storage.get("fixed_color"))
    )

    state["lights"] = lights_state
    return state


@user_passes_test(user_manager.has_controls)
def index(request: WSGIRequest) -> HttpResponse:
    """Renders the /lights page. During voting, only privileged users can access this."""
    from core import urls

    context = base.context(request)
    context["urls"] = urls.lights_paths
    # programs that have a strip_color or ring_color function are color programs
    # programs that have a draw function are screen programs
    context["color_program_names"] = ["Disabled", "Fixed", "Rainbow", "Rave"]
    context["screen_program_names"] = ["Disabled", "Circular"]
    return render(request, "lights.html", context)


def update_state() -> None:
    """Sends an update event to all connected clients."""
    send_state(state_dict())



"""This module provides common functionality for all pages on the site."""

import os
import random
from typing import Dict, Any

from django.conf import settings as conf
from django.core.handlers.wsgi import WSGIRequest
from django.db import transaction
from django.http import HttpResponseBadRequest
from django.http import HttpResponseRedirect
from django.http.response import HttpResponse, JsonResponse
from django.shortcuts import render
from django.urls import reverse

import core.settings.storage as storage
import core.lights.controller as lights_controller
from core import models

from core import user_manager
from core import redis
from core.state_handler import send_state


def _get_random_hashtag() -> str:
    active_hashtags = models.Tag.objects.filter(active=True)
    if active_hashtags.count() == 0:
        return "no hashtags present :("
    index = random.randint(0, active_hashtags.count() - 1)
    hashtag = active_hashtags[index]
    return hashtag.text


def _get_apk_link() -> str:
    local_apk = os.path.join(conf.STATIC_FILES, "apk/shareberry.apk")
    if os.path.isfile(local_apk):
        return os.path.join(conf.STATIC_URL, "apk/shareberry.apk")
    return "https://github.com/raveberry/shareberry/releases/latest/download/shareberry.apk"


def _increment_counter() -> int:
    with transaction.atomic():
        counter = models.Counter.objects.get_or_create(id=1, defaults={"value": 0})[0]
        counter.value += 1
        counter.save()
    update_state()
    return counter.value


def context(request: WSGIRequest) -> Dict[str, Any]:
    """Returns the base context that is needed on every page.
    Increments the visitors counter."""
    from core import urls

    _increment_counter()
    return {
        "base_urls": urls.base_paths,
        "voting_enabled": storage.get("voting_enabled"),
        "hashtag": _get_random_hashtag(),
        "demo": conf.DEMO,
        "controls_enabled": user_manager.has_controls(request.user),
        "is_admin": user_manager.is_admin(request.user),
        "apk_link": _get_apk_link(),
        "local_enabled": storage.get("local_enabled"),
        "youtube_enabled": storage.get("youtube_enabled"),
        "spotify_enabled": storage.get("spotify_enabled"),
        "soundcloud_enabled": storage.get("soundcloud_enabled"),
        "jamendo_enabled": storage.get("jamendo_enabled"),
        "streaming_enabled": conf.DOCKER_ICECAST or storage.get("output") == "icecast",
    }


def state_dict() -> Dict[str, Any]:
    """This function constructs a base state dictionary with website wide state.
    Pages sending states extend this state dictionary."""
    return {
        "partymode": user_manager.partymode_enabled(),
        "users": user_manager.get_count(),
        "visitors": models.Counter.objects.get_or_create(id=1, defaults={"value": 0})[
            0
        ].value,
        "lightsEnabled": redis.get("lights_active"),
        "playbackError": redis.get("playback_error"),
        "alarm": redis.get("alarm_playing"),
        "defaultPlatform": "spotify" if storage.get("spotify_enabled") else "youtube",
    }


def no_stream(request: WSGIRequest) -> HttpResponse:
    """Renders the /stream page. If this is reached, there is no stream active."""
    return render(request, "no_stream.html", context(request))


def submit_hashtag(request: WSGIRequest) -> HttpResponse:
    """Add the given hashtag to the database."""
    hashtag = request.POST.get("hashtag")
    if hashtag is None or len(hashtag) == 0:
        return HttpResponseBadRequest()

    if hashtag[0] != "#":
        hashtag = "#" + hashtag
    models.Tag.objects.create(text=hashtag, active=storage.get("hashtags_active"))

    return HttpResponse()


def logged_in(request: WSGIRequest) -> HttpResponse:
    """This endpoint is visited after every login.
    Redirect the admin to the settings and everybody else to the musiq page."""
    if user_manager.is_admin(request.user):
        return HttpResponseRedirect(reverse("settings"))
    return HttpResponseRedirect(reverse("musiq"))


def set_lights_shortcut(request: WSGIRequest) -> None:
    """Request endpoint for the lights shortcut.
    Situated in base because the dropdown is accessible from every page."""
    return lights_controller.set_lights_shortcut(request)


def upgrade_available(_request: WSGIRequest) -> HttpResponse:
    """Checks whether newer Raveberry version is available."""
    from core.settings import system

    latest_version = system.fetch_latest_version()
    current_version = conf.VERSION
    if latest_version and latest_version != current_version:
        return JsonResponse(True, safe=False)
    return JsonResponse(False, safe=False)


def update_state() -> None:
    """Sends an update event to all connected clients."""
    send_state(state_dict())



"""This module defines the routing for realtime websocket requests."""

from django.conf.urls import url
from core import state_handler

WEBSOCKET_URLPATTERNS = [url(r"^state/$", state_handler.StateConsumer)]



"""This module provides functionality to interface with Redis."""
from ast import literal_eval
from typing import Any, Union, List, Dict

from django.conf import settings as conf
from redis import Redis

# locks:
# player_lock:  controlling mopidy api accesses
# lights_lock:  ensures lights settings are not changed during device updates

# channels
# lights_settings_changed

# values:
# maps key to default and type of value
defaults = {
    # playback
    "playing": False,
    "playback_error": False,
    "stop_playback_loop": False,
    "alarm_playing": False,
    "alarm_requested": False,
    "backup_playing": False,
    # lights
    "lights_active": False,
    "ring_initialized": False,
    "wled_initialized": False,
    "strip_initialized": False,
    "screen_initialized": False,
    # settings
    "has_internet": False,
    "youtube_available": False,
    "spotify_available": False,
    "soundcloud_available": False,
    "jamendo_available": False,
    "library_scan_progress": "0 / 0 / 0",
    "bluetoothctl_active": False,
    "bluetooth_devices": [],
    # user manager
    "last_user_count_update": 0.0,
    "last_requests": {},
}

redis_connection = Redis(
    host=conf.REDIS_HOST, port=conf.REDIS_PORT, decode_responses=True
)


def start() -> None:
    redis_connection.flushdb()


lock = redis_connection.lock
pubsub = redis_connection.pubsub
publish = redis_connection.publish


def get(key: str) -> Union[bool, int, float, str, List, Dict]:
    """This method returns the value for the given :param key: from redis.
    Vaules of non-existing keys are set to their respective default value."""
    # values are stored as string in redis
    # cast the value to its respective type, defined by the default value, before returning it
    default = defaults[key]
    value = redis_connection.get(key)
    if value is None:
        return default
    if type(default) == bool:
        # bool("False") does not return False -> special case for bool
        return value == "True"
    elif type(default) in (list, dict):
        # evaluate the stored literal
        return literal_eval(value)
    return type(default)(value)


def set(key: str, value: Any) -> None:
    """This method sets the value for the given :param key: to the given :param value:."""
    redis_connection.set(key, str(value))


class Event:
    """A class that provides basic functionality similar to threading.Event using redis."""

    def __init__(self, name: str) -> None:
        self.name = name
        self.is_set = False
        self.lock = redis_connection.lock(f"{self.name}_lock")

    def wait(self) -> None:
        """Blocks until the event is set."""
        with self.lock:
            is_set = self.is_set
        if is_set:
            return
        p = redis_connection.pubsub(ignore_subscribe_messages=True)
        p.subscribe(self.name)
        next(p.listen())
        p.close()

    def set(self) -> None:
        """Set the event and wake up all waiting threads."""
        with self.lock:
            self.is_set = True
            redis_connection.publish(self.name, "")

    def clear(self) -> None:
        """Clear this Event, allowing threads to wait for it."""
        with self.lock:
            self.is_set = False



"""This module contains all Youtube related code."""

from __future__ import annotations

import errno
import json
import logging
import os
import pickle
import shutil
import subprocess
from contextlib import contextmanager
from typing import Any, Dict, List, Optional, TYPE_CHECKING, Iterator, cast
from urllib.parse import parse_qs
from urllib.parse import urlparse

import requests
import youtube_dl
from django.conf import settings
from django.http.response import HttpResponse

import core.musiq.song_utils as song_utils
import core.settings.storage as storage
from core.musiq import musiq
from core.musiq.song_provider import SongProvider
from core.musiq.playlist_provider import PlaylistProvider

if TYPE_CHECKING:
    from core.musiq.song_utils import Metadata


@contextmanager
def youtube_session() -> Iterator[requests.Session]:
    """This context opens a requests session and loads the youtube cookies file."""
    session = requests.session()

    pickle_file = os.path.join(settings.BASE_DIR, "config/youtube_cookies.pickle")

    session = requests.session()
    # Have youtube-dl deal with consent cookies etc to setup a valid session
    extractor = youtube_dl.extractor.youtube.YoutubeIE()
    extractor._downloader = youtube_dl.YoutubeDL()
    extractor.initialize()
    session.cookies.update(extractor._downloader.cookiejar)

    try:
        if os.path.getsize(pickle_file) > 0:
            with open(pickle_file, "rb") as f:
                session.cookies.update(pickle.load(f))
    except FileNotFoundError:
        pass

    headers = {"User-Agent": youtube_dl.utils.random_user_agent()}
    session.headers.update(headers)
    yield session

    with open(pickle_file, "wb") as f:
        pickle.dump(session.cookies, f)


class YoutubeDLLogger:
    """This logger class is used to log process of youtube-dl downloads."""

    @classmethod
    def debug(cls, msg: str) -> None:
        """This method is called if youtube-dl does debug level logging."""
        logging.debug(msg)

    @classmethod
    def warning(cls, msg: str) -> None:
        """This method is called if youtube-dl does warning level logging."""
        logging.debug(msg)

    @classmethod
    def error(cls, msg: str) -> None:
        """This method is called if youtube-dl does error level logging."""
        logging.error(msg)


class Youtube:
    """This class contains code for both the song and playlist provider"""

    atomicparsley_available = shutil.which("AtomicParsley") is not None

    @staticmethod
    def get_ydl_opts() -> Dict[str, Any]:
        """This method returns a dictionary containing sensible defaults for youtube-dl options.
        It is roughly equivalent to the following command:
        youtube-dl --format bestaudio[ext=m4a]/best[ext=m4a] --output '%(id)s.%(ext)s' \
            --no-playlist --no-cache-dir --write-thumbnail --default-search auto \
            --add-metadata --embed-thumbnail
        """
        postprocessors = [{"key": "FFmpegMetadata"}]
        if Youtube.atomicparsley_available:
            postprocessors.append(
                {
                    "key": "EmbedThumbnail",
                    # overwrite any thumbnails already present
                    "already_have_thumbnail": True,
                }
            )
        return {
            "format": "bestaudio[ext=m4a]/best[ext=m4a]",
            "outtmpl": os.path.join(settings.SONGS_CACHE_DIR, "%(id)s.%(ext)s"),
            "noplaylist": True,
            "cachedir": False,
            "no_color": True,
            "writethumbnail": True,
            "default_search": "auto",
            "postprocessors": postprocessors,
            "logger": YoutubeDLLogger(),
        }

    @staticmethod
    def _get_initial_data(html: str) -> Dict[str, Any]:
        for line in html.split("\n"):
            line = line.strip()
            before = "var ytInitialData = "
            after = ";</"
            if before in line:
                # extract json
                line = line[line.index(before) + len(before) :]
                initial_data = line[: line.index(after)]
                return json.loads(initial_data)
        raise ValueError("Could not parse initial data from html")

    @staticmethod
    def get_search_suggestions(query: str) -> List[str]:
        """Returns a list of suggestions for the given query from Youtube."""
        with youtube_session() as session:
            params = {
                "client": "youtube",
                "q": query[:100],  # queries longer than 100 characters are not accepted
                "xhr": "t",  # this makes the response be a json file
            }
            response = session.get(
                "https://clients1.google.com/complete/search", params=params
            )
        suggestions = json.loads(response.text)
        # first entry is the query, the second one contains the suggestions
        suggestions = suggestions[1]
        # suggestions are given as tuples
        # extract the string and skip the query if it occurs identically
        suggestions = [
            entry[0]
            for entry in suggestions
            if entry[0] != query and not song_utils.is_forbidden(entry[0])
        ]
        return suggestions


class YoutubeSongProvider(SongProvider, Youtube):
    """This class handles songs from Youtube."""

    @staticmethod
    def get_id_from_external_url(url: str) -> str:
        return parse_qs(urlparse(url).query)["v"][0]

    def __init__(self, query: Optional[str], key: Optional[int]) -> None:
        self.type = "youtube"
        super().__init__(query, key)
        self.info_dict: Dict[str, Any] = {}
        self.ydl_opts = Youtube.get_ydl_opts()

    def check_cached(self) -> bool:
        if not self.id:
            return False
        return os.path.isfile(self._get_path())

    def check_available(self) -> bool:

        # directly use the search extractors entry function so we can process each result
        # as soon as it's available instead of waiting for all of them
        extractor = youtube_dl.extractor.youtube.YoutubeSearchIE()
        extractor._downloader = youtube_dl.YoutubeDL(self.ydl_opts)
        extractor.initialize()
        for entry in extractor._entries(self.query, 50):
            if song_utils.is_forbidden(entry["title"]):
                continue
            try:
                with youtube_dl.YoutubeDL(self.ydl_opts) as ydl:
                    self.info_dict = ydl.extract_info(entry["id"], download=False)
                break
            except (
                youtube_dl.utils.ExtractorError,
                youtube_dl.utils.DownloadError,
            ) as e:
                logging.warning("error during availability check for %s:", entry["id"])
                logging.warning(e)
        else:
            self.error = "No songs found"
            return False

        self.id = self.info_dict["id"]

        return self.check_not_too_large(self.info_dict["filesize"])

    def _download(self) -> bool:
        error = None
        location = None

        try:
            with youtube_dl.YoutubeDL(self.ydl_opts) as ydl:
                ydl.download([self.get_external_url()])

            location = self._get_path()
            base = os.path.splitext(location)[0]
            thumbnail = base + ".jpg"
            try:
                os.remove(thumbnail)
            except FileNotFoundError:
                logging.info("tried to delete %s but does not exist", thumbnail)

            try:
                # tag the file with replaygain to perform volume normalization
                subprocess.call(
                    ["rganalysis", location],
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                )
            except OSError as e:
                if e.errno == errno.ENOENT:
                    pass  # the rganalysis package was not found. Skip normalization
                else:
                    raise

        except youtube_dl.utils.DownloadError as e:
            error = e

        if error is not None or location is None:
            logging.error("accessible video could not be downloaded: %s", self.id)
            logging.error("location: %s", location)
            logging.error(error)
            return False
        return True

    def make_available(self) -> bool:
        if not os.path.isfile(self._get_path()):
            musiq.update_state()
            # only download the file if it was not already downloaded
            return self._download()
        return True

    def get_metadata(self) -> "Metadata":
        if not self.id:
            raise ValueError()
        metadata = song_utils.get_metadata(self._get_path())

        if not metadata["title"]:
            metadata["title"] = metadata["external_url"]
        metadata["internal_url"] = self.get_internal_url()
        metadata["external_url"] = "https://www.youtube.com/watch?v=" + self.id
        metadata["stream_url"] = None

        return metadata

    def _get_path(self) -> str:
        if not self.id:
            raise ValueError()
        return song_utils.get_path(self.id + ".m4a")

    def get_internal_url(self) -> str:
        return "file://" + self._get_path()

    def get_external_url(self) -> str:
        if not self.id:
            raise ValueError()
        return "https://www.youtube.com/watch?v=" + self.id

    def get_suggestion(self) -> str:
        with youtube_session() as session:
            response = session.get(self.get_external_url())

        initial_data = Youtube._get_initial_data(response.text)

        path = [
            "contents",
            "twoColumnWatchNextResults",
            "autoplay",
            "autoplay",
            "sets",
            0,
            "autoplayVideo",
            "commandMetadata",
            "webCommandMetadata",
            "url",
        ]
        url = initial_data
        for step in path:
            url = url[cast(str, step)]
        # discard everything after the first v= parameter
        return "https://www.youtube.com" + cast(str, url).split("&")[0]

    def request_radio(self, request_ip: str) -> HttpResponse:
        if not self.id:
            raise ValueError()
        radio_id = "RD" + self.id

        provider = YoutubePlaylistProvider("", None)
        provider.id = radio_id
        provider.request("", archive=False, manually_requested=False)
        return HttpResponse("queueing radio (might take some time)")


class YoutubePlaylistProvider(PlaylistProvider, Youtube):
    """This class handles Youtube Playlists."""

    @staticmethod
    def get_id_from_external_url(url: str) -> Optional[str]:
        try:
            list_id = parse_qs(urlparse(url).query)["list"][0]
        except KeyError:
            return None
        return list_id

    def __init__(self, query: Optional[str], key: Optional[int]) -> None:
        self.type = "youtube"
        super().__init__(query, key)
        self.ydl_opts = Youtube.get_ydl_opts()
        del self.ydl_opts["noplaylist"]
        self.ydl_opts["extract_flat"] = True

    def is_radio(self) -> bool:
        if not self.id:
            raise ValueError()
        return self.id.startswith("RD")

    def search_id(self) -> Optional[str]:
        with youtube_session() as session:
            params = {
                "search_query": self.query,
                # this is the value that youtube uses to filter for playlists only
                "sp": "EgQQA1AD",
            }
            response = session.get("https://www.youtube.com/results", params=params)

        initial_data = Youtube._get_initial_data(response.text)

        path = [
            "contents",
            "twoColumnSearchResultsRenderer",
            "primaryContents",
            "sectionListRenderer",
            "contents",
        ]
        section_renderers = initial_data
        for step in path:
            section_renderers = section_renderers[step]

        list_id = None
        for section_renderer in cast(List[Dict[str, Any]], section_renderers):
            search_results = section_renderer["itemSectionRenderer"]["contents"]

            try:
                list_id = next(
                    res["playlistRenderer"]["playlistId"]
                    for res in search_results
                    if "playlistRenderer" in res
                )
                break
            except StopIteration:
                # the search result did not contain the list id
                pass

        return list_id

    def fetch_metadata(self) -> bool:
        # in case of a radio playlist, restrict the number of songs that are downloaded
        if self.is_radio():
            self.ydl_opts["playlistend"] = storage.get("max_playlist_items")
            # radios are not viewable with the /playlist?list= url,
            # create a video watch url with the radio list
            query_url = (
                "https://www.youtube.com/watch?v=" + self.id[2:] + "&list=" + self.id
            )
        else:
            # if only given the id, youtube-dl returns an info dict resolving this id to a url.
            # we want to receive the playlist entries directly, so we query the playlist url
            query_url = "https://www.youtube.com/playlist?list=" + self.id

        try:
            with youtube_dl.YoutubeDL(self.ydl_opts) as ydl:
                info_dict = ydl.extract_info(query_url, download=False)
        except (youtube_dl.utils.ExtractorError, youtube_dl.utils.DownloadError) as e:
            self.error = e
            return False

        if info_dict["_type"] != "playlist" or "entries" not in info_dict:
            # query was not a playlist url -> search for the query
            assert False

        assert self.id == info_dict["id"]
        if "title" in info_dict:
            self.title = info_dict["title"]
        for entry in info_dict["entries"]:
            self.urls.append("https://www.youtube.com/watch?v=" + entry["id"])
        assert self.key is None

        return True



"""This module contains the base class of all song providers."""

import logging
from typing import Optional, Type, TYPE_CHECKING

from django.db import transaction
from django.db.models.expressions import F
from django.http.response import HttpResponse

import core.settings.storage as storage
from core.models import ArchivedSong, QueuedSong, ArchivedQuery, RequestLog
from core.musiq import song_utils as song_utils, playback
from core.musiq import musiq
from core.musiq.music_provider import MusicProvider, WrongUrlError

if TYPE_CHECKING:
    from core.musiq.song_utils import Metadata


class SongProvider(MusicProvider):
    """The base class for all single song providers."""

    @staticmethod
    def get_id_from_external_url(url: str) -> str:
        """Constructs and returns the external id based on the given url."""
        raise NotImplementedError()

    @staticmethod
    def create(
        query: Optional[str] = None,
        key: Optional[int] = None,
        external_url: Optional[str] = None,
    ) -> "SongProvider":
        """Factory method to create a song provider.
        Either (query and key) or external url need to be specified.
        Detects the type of provider needed and returns one of corresponding type."""
        if key is not None:
            if query is None:
                logging.error(
                    "archived song requested but no query given (key %s)", key
                )
                raise ValueError()
            try:
                archived_song = ArchivedSong.objects.get(id=key)
            except ArchivedSong.DoesNotExist:
                logging.error("archived song requested for nonexistent key %s", key)
                raise ValueError()
            external_url = archived_song.url
        if external_url is None:
            raise ValueError(
                "external_url was provided and could not be inferred from remaining attributes."
            )
        provider_class: Optional[Type[SongProvider]] = None
        url_type = song_utils.determine_url_type(external_url)
        if url_type == "local":
            from core.musiq.localdrive import LocalSongProvider

            provider_class = LocalSongProvider
        elif storage.get("youtube_enabled") and url_type == "youtube":
            from core.musiq.youtube import YoutubeSongProvider

            provider_class = YoutubeSongProvider
        elif storage.get("spotify_enabled") and url_type == "spotify":
            from core.musiq.spotify import SpotifySongProvider

            provider_class = SpotifySongProvider
        elif storage.get("soundcloud_enabled") and url_type == "soundcloud":
            from core.musiq.soundcloud import SoundcloudSongProvider

            provider_class = SoundcloudSongProvider
        elif storage.get("jamendo_enabled") and url_type == "jamendo":
            from core.musiq.jamendo import JamendoSongProvider

            provider_class = JamendoSongProvider
        if not provider_class:
            raise NotImplementedError(f"No provider for given song: {external_url}")
        if not query and external_url:
            query = external_url
        provider = provider_class(query, key)
        return provider

    def __init__(self, query: Optional[str], key: Optional[int]) -> None:
        super().__init__(query, key)
        self.ok_message = "song queued"
        self.queued_song: Optional[QueuedSong] = None

        if query:
            url_type = song_utils.determine_url_type(query)
            if url_type not in (self.type, "unknown"):
                raise WrongUrlError(
                    f"Tried to create a {self.type} provider with: {query}"
                )

    def _get_path(self) -> str:
        raise NotImplementedError()

    def get_internal_url(self) -> str:
        """Returns the internal url based on this object's id."""
        raise NotImplementedError()

    def get_external_url(self) -> str:
        """Returns the external url based on this object's id."""
        raise NotImplementedError()

    def extract_id(self) -> Optional[str]:
        if self.key is not None:
            try:
                archived_song = ArchivedSong.objects.get(id=self.key)
                return self.__class__.get_id_from_external_url(archived_song.url)
            except ArchivedSong.DoesNotExist:
                return None
        if self.query is not None:
            url_type = song_utils.determine_url_type(self.query)
            if url_type == "local":
                from core.musiq.localdrive import LocalSongProvider

                return LocalSongProvider.get_id_from_external_url(self.query)
            if storage.get("youtube_enabled") and url_type == "youtube":
                from core.musiq.youtube import YoutubeSongProvider

                return YoutubeSongProvider.get_id_from_external_url(self.query)
            if storage.get("spotify_enabled") and url_type == "spotify":
                from core.musiq.spotify import SpotifySongProvider

                return SpotifySongProvider.get_id_from_external_url(self.query)
            if storage.get("soundcloud_enabled") and url_type == "soundcloud":
                from core.musiq.soundcloud import SoundcloudSongProvider

                return SoundcloudSongProvider.get_id_from_external_url(self.query)
            if storage.get("jamendo_enabled") and url_type == "jamendo":
                from core.musiq.jamendo import JamendoSongProvider

                return JamendoSongProvider.get_id_from_external_url(self.query)
            try:
                archived_song = ArchivedSong.objects.get(url=self.query)
                return self.__class__.get_id_from_external_url(archived_song.url)
            except ArchivedSong.DoesNotExist:
                return None
        logging.error("Can not extract id because neither key nor query are known")
        return None

    def enqueue_placeholder(self, manually_requested) -> None:
        metadata: Metadata = {
            "artist": "",
            "title": self.query or self.get_external_url(),
            "duration": -1,
            "internal_url": None,
            "external_url": self.get_external_url(),
            "stream_url": None,
        }
        initial_votes = 1 if manually_requested else 0
        self.queued_song = playback.queue.enqueue(
            metadata, manually_requested, votes=initial_votes
        )

    def remove_placeholder(self) -> None:
        assert self.queued_song
        self.queued_song.delete()

    def check_cached(self) -> bool:
        raise NotImplementedError()

    def check_not_too_large(self, size: Optional[float]) -> bool:
        max_size = storage.get("max_download_size") * 1024 * 1024
        if (
            max_size != 0
            and not self.check_cached()
            and (size is not None and size > max_size)
        ):
            self.error = "Song too long"
            return False
        return True

    def check_available(self) -> bool:
        raise NotImplementedError()

    def make_available(self) -> bool:
        return True

    def persist(self, request_ip: str, archive: bool = True) -> None:
        metadata = self.get_metadata()

        # Increase counter of song/playlist
        with transaction.atomic():
            queryset = ArchivedSong.objects.filter(url=metadata["external_url"])
            if queryset.count() == 0:
                initial_counter = 1 if archive else 0
                archived_song = ArchivedSong.objects.create(
                    url=metadata["external_url"],
                    artist=metadata["artist"],
                    title=metadata["title"],
                    counter=initial_counter,
                )
            else:
                if archive:
                    queryset.update(counter=F("counter") + 1)
                archived_song = queryset.get()

            if archive:
                ArchivedQuery.objects.get_or_create(
                    song=archived_song, query=self.query
                )

        if storage.get("logging_enabled") and request_ip:
            RequestLog.objects.create(song=archived_song, address=request_ip)

    def enqueue(self) -> None:
        assert self.queued_song
        if not playback.queue.filter(id=self.queued_song.id).exists():
            # this song was already deleted, do not enqueue
            return

        metadata = self.get_metadata()

        self.queued_song.internal_url = metadata["internal_url"]
        self.queued_song.external_url = metadata["external_url"]
        self.queued_song.stream_url = metadata["stream_url"]
        self.queued_song.artist = metadata["artist"]
        self.queued_song.title = metadata["title"]
        self.queued_song.duration = metadata["duration"]
        # make sure not to overwrite the index as it may have changed in the meantime
        self.queued_song.save(
            update_fields=[
                "artist",
                "title",
                "duration",
                "internal_url",
                "external_url",
                "stream_url",
            ]
        )

        musiq.update_state()
        playback.queue_changed.set()

    def get_suggestion(self) -> str:
        """Returns the external url of a suggested song based on this one."""
        raise NotImplementedError()

    def get_metadata(self) -> "Metadata":
        """Returns a dictionary of this song's metadata."""
        raise NotImplementedError()

    def request_radio(self, request_ip) -> HttpResponse:
        """Enqueues a playlist of songs based on this one."""
        raise NotImplementedError()



"""This module handles all settings related to sound output."""
from __future__ import annotations

import re
import subprocess
import time
from threading import Thread
from typing import Optional

from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse
from django.http import HttpResponseBadRequest
from django.http import JsonResponse

from core import util, redis
from core.celery import app
from core.models import CurrentSong
from core.settings import storage, system, settings
from core.settings.settings import control

# to control that only one bluetoothctl process is active at a time
# we use a redis variable instead of a redis lock
# we need to release the lock(=kill the process) in a different request than the one that started it
# this made using the lock complicated, and only one admin should be using the page at once anyway

# bluetooth_devices: List[Dict[str, str]] = []


@control
def set_backup_stream(request: WSGIRequest) -> None:
    """Sets the given internet stream as backup stream."""
    stream = request.POST.get("value")
    storage.set("backup_stream", stream)


def _get_bluetoothctl_line(bluetoothctl: subprocess.Popen[bytes]) -> str:
    assert bluetoothctl.stdout
    line = bluetoothctl.stdout.readline().decode()
    ansi_escape = re.compile(r"\x1B\[[0-?]*[ -/]*[@-~]")
    line = ansi_escape.sub("", line)
    line = line.strip()
    return line


def _start_bluetoothctl() -> Optional[subprocess.Popen[bytes]]:
    if redis.get("bluetoothctl_active"):
        return None
    redis.set("bluetoothctl_active", True)
    bluetoothctl = subprocess.Popen(
        ["bluetoothctl"], stdin=subprocess.PIPE, stdout=subprocess.PIPE
    )
    return bluetoothctl


def _stop_bluetoothctl(bluetoothctl: subprocess.Popen[bytes]) -> None:
    assert bluetoothctl.stdin
    bluetoothctl.stdin.close()
    bluetoothctl.wait()
    redis.set("bluetoothctl_active", False)


@app.task
def _scan_bluetooth() -> None:
    bluetoothctl = _start_bluetoothctl()
    redis.set("bluetooth_devices", [])
    assert bluetoothctl.stdin

    bluetoothctl.stdin.write(b"devices\n")
    bluetoothctl.stdin.write(b"scan on\n")
    bluetoothctl.stdin.flush()
    while True:
        line = _get_bluetoothctl_line(bluetoothctl)
        if not line:
            break
        # match old devices
        match = re.match(r"Device (\S*) (.*)", line)
        # match newly scanned devices
        # We need the '.*' at the beginning of the line to account for control sequences
        if not match:
            match = re.match(r".*\[NEW\] Device (\S*) (.*)", line)
        if match:
            address = match.group(1)
            name = match.group(2)
            # filter unnamed devices
            # devices named after their address are no speakers
            if re.match("[A-Z0-9][A-Z0-9](-[A-Z0-9][A-Z0-9]){5}", name):
                continue
            bluetooth_devices = redis.get("bluetooth_devices")
            bluetooth_devices.append({"address": address, "name": name})
            redis.set("bluetooth_devices", bluetooth_devices)
            settings.update_state()


@control
def set_bluetooth_scanning(request: WSGIRequest) -> HttpResponse:
    """Enables scanning of bluetooth devices."""
    enabled = request.POST.get("value") == "true"
    if enabled:
        if redis.get("bluetoothctl_active"):
            return HttpResponseBadRequest("Already Scanning")

        _scan_bluetooth.delay()
        return HttpResponse("Started scanning")

    if not redis.get("bluetoothctl_active"):
        return HttpResponseBadRequest("Currently not scanning")
    # this is another request, so we don't have a handle of the current bluetoothctl process
    # terminate the process by name and release the lock
    subprocess.call("pkill bluetoothctl".split())
    redis.set("bluetoothctl_active", False)
    return HttpResponse("Stopped scanning")


@control
def connect_bluetooth(request: WSGIRequest) -> HttpResponse:
    """Connect to a given bluetooth device."""
    address = request.POST.get("address")
    if address is None or address == "":
        return HttpResponseBadRequest("No device selected")
    bluetoothctl = _start_bluetoothctl()
    if not bluetoothctl:
        return HttpResponseBadRequest("Stop scanning before connecting")

    assert bluetoothctl.stdin
    error = ""

    # A Function that acts as a timeout for unexpected errors (or timeouts)
    def _timeout() -> None:
        nonlocal error
        time.sleep(20)
        error = "Timed out"
        if bluetoothctl is not None:
            _stop_bluetoothctl(bluetoothctl)

    Thread(target=_timeout).start()

    # Sometimes, pairing hangs forever. Since connecting alone is enough, skip pairing.
    # bluetoothctl.stdin.write(b"pair " + address.encode() + b"\n")
    # bluetoothctl.stdin.flush()
    # while True:
    #     line = _get_bluetoothctl_line(bluetoothctl)
    #     if not line:
    #         break
    #     if re.match(".*Device " + address + " not available", line):
    #         error = "Device unavailable"
    #         break
    #     if re.match(".*Failed to pair: org.bluez.Error.AlreadyExists", line):
    #         break
    #     if re.match(".*Pairing successful", line):
    #         break

    # if error:
    #     _stop_bluetoothctl()
    #     return HttpResponseBadRequest(error)

    bluetoothctl.stdin.write(b"connect " + address.encode() + b"\n")
    bluetoothctl.stdin.flush()
    while True:
        line = _get_bluetoothctl_line(bluetoothctl)
        if not line:
            break
        if re.match(".*Device " + address + " not available", line):
            error = "Device unavailable"
            break
        if re.match(".*Failed to connect: org.bluez.Error.Failed", line):
            error = "Connect Failed"
            break
        if re.match(".*Failed to connect: org.bluez.Error.InProgress", line):
            error = "Connect in progress"
            break
        if re.match(".*Connection successful", line):
            break
    # trust the device to automatically reconnect when it is available again
    bluetoothctl.stdin.write(b"trust " + address.encode() + b"\n")
    bluetoothctl.stdin.flush()

    _stop_bluetoothctl(bluetoothctl)
    if error:
        return HttpResponseBadRequest(error)

    return HttpResponse("Connected. Set output device to activate.")


@control
def disconnect_bluetooth(request: WSGIRequest) -> HttpResponse:
    """Untrusts a given bluetooth device to prevent automatic reconnects.
    Does not unpair or remove the device."""
    address = request.POST.get("address")
    if address is None or address == "":
        return HttpResponseBadRequest("No device selected")
    bluetoothctl = _start_bluetoothctl()
    if not bluetoothctl:
        return HttpResponseBadRequest("Stop scanning before disconnecting")
    assert bluetoothctl.stdin
    error = ""

    bluetoothctl.stdin.write(b"untrust " + address.encode() + b"\n")
    bluetoothctl.stdin.flush()
    while True:
        line = _get_bluetoothctl_line(bluetoothctl)
        if not line:
            break
        if re.match(".*Device " + address + " not available", line):
            error = "Device unavailable"
            break
        if re.match(".*untrust succeeded", line):
            break

    _stop_bluetoothctl(bluetoothctl)
    if error:
        return HttpResponseBadRequest(error)
    return HttpResponse("Disconnected")


@control
def set_feed_cava(request: WSGIRequest) -> None:
    """Enables or disables whether mopidy should output to the cava fake device."""
    enabled = request.POST.get("value") == "true"
    storage.set("feed_cava", enabled)
    # update mopidy config to apply the change
    system.update_mopidy_config("pulse")


@control
def list_outputs(_request: WSGIRequest) -> JsonResponse:
    """Returns a list of all sound output devices currently available."""
    output = subprocess.check_output(
        "pactl list short sinks".split(),
        env={"PULSE_SERVER": "127.0.0.1"},
        universal_newlines=True,
    )
    tokenized_lines = [line.split() for line in output.splitlines()]

    sinks = ["fakesink", "icecast", "snapcast"]
    sinks.extend([sink[1] for sink in tokenized_lines if len(sink) >= 2])

    return JsonResponse(sinks, safe=False)


def _set_output(output: str) -> HttpResponse:
    icecast_installed = util.service_installed("icecast2")
    snapcast_installed = util.service_installed("snapserver")

    if output == "fakesink":
        mopidy_output = "fakesink"
    elif output == "icecast":
        if not icecast_installed:
            return HttpResponseBadRequest("Please install icecast2")

        subprocess.call(["sudo", "/usr/local/sbin/raveberry/enable_icecast"])
        mopidy_output = "icecast"
    elif output == "snapcast":
        if not snapcast_installed:
            return HttpResponseBadRequest("Please install snapserver")

        subprocess.call(["sudo", "/usr/local/sbin/raveberry/enable_snapcast"])
        mopidy_output = "snapcast"
    else:
        try:
            subprocess.run(
                ["pactl", "set-default-sink", output],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.PIPE,
                env={"PULSE_SERVER": "127.0.0.1"},
                check=True,
            )
            mopidy_output = "pulse"
        except subprocess.CalledProcessError as e:
            return HttpResponseBadRequest(e.stderr)

    if icecast_installed and output != "icecast":
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/disable_icecast"])
    if snapcast_installed and output != "snapcast":
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/disable_snapcast"])

    system.update_mopidy_config(mopidy_output)

    return HttpResponse(
        "Output was set. Restarting the current song might be necessary."
    )


@control
def set_output(request: WSGIRequest) -> HttpResponse:
    """Sets the given device as default output device."""
    output = request.POST.get("value")
    if not output:
        return HttpResponseBadRequest("No output selected")

    if output == request.POST.get("output"):
        return HttpResponseBadRequest("Output unchanged")

    storage.set("output", output)

    return _set_output(output)


@control
def delete_current_song(_request: WSGIRequest) -> HttpResponse:
    """Force skips the current song by deleting it from the database."""
    try:
        current_song = CurrentSong.objects.get()
        current_song.delete()
        return HttpResponse()
    except CurrentSong.DoesNotExist:
        return HttpResponseBadRequest("No song playing")


@control
def restart_player(_request: WSGIRequest) -> None:
    """Restarts mopidy."""
    system._restart_mopidy()



"""This module contains everything related to the settings and configuration of the server."""

from __future__ import annotations

import os
import subprocess
from functools import wraps

import yaml
from typing import Dict, Any, Callable, Optional

from django.conf import settings as conf
from django.contrib.auth.decorators import user_passes_test
from django.core.handlers.wsgi import WSGIRequest

# JsonResponse needs to be imported here so types can resolved through the decorator?
from django.http import HttpResponse, HttpResponseForbidden
from django.shortcuts import render

from core import user_manager, base, redis
from core.settings.storage import get
from core.state_handler import send_state


def control(
    func: Callable[[WSGIRequest], Optional[HttpResponse]]
) -> Callable[[WSGIRequest], HttpResponse]:
    """A decorator that makes sure that only the admin changes a setting."""

    def _decorator(request: WSGIRequest) -> HttpResponse:
        if not user_manager.is_admin(request.user):
            return HttpResponseForbidden()
        response = func(request)
        update_state()
        if response is not None:
            return response
        return HttpResponse()

    return wraps(func)(_decorator)


def state_dict() -> Dict[str, Any]:
    state = base.state_dict()

    settings_state = {}
    settings_state["votingEnabled"] = get("voting_enabled")
    settings_state["newMusicOnly"] = get("new_music_only")
    settings_state["loggingEnabled"] = get("logging_enabled")
    settings_state["embedStream"] = get("embed_stream")
    settings_state["dynamicEmbeddedStream"] = get("dynamic_embedded_stream")
    settings_state["hashtagsActive"] = get("hashtags_active")
    settings_state["onlineSuggestions"] = get("online_suggestions")
    settings_state["numberOfSuggestions"] = get("number_of_suggestions")
    settings_state["peopleToParty"] = get("people_to_party")
    settings_state["alarmProbability"] = get("alarm_probability")
    settings_state["buzzerCooldown"] = get("buzzer_cooldown")
    settings_state["downvotesToKick"] = get("downvotes_to_kick")
    settings_state["maxDownloadSize"] = get("max_download_size")
    settings_state["additionalKeywords"] = get("additional_keywords")
    settings_state["forbiddenKeywords"] = get("forbidden_keywords")
    settings_state["maxPlaylistItems"] = get("max_playlist_items")
    settings_state["maxQueueLength"] = get("max_queue_length")
    settings_state["hasInternet"] = redis.get("has_internet")

    settings_state["youtubeEnabled"] = get("youtube_enabled")
    settings_state["youtubeSuggestions"] = get("youtube_suggestions")

    settings_state["spotifyEnabled"] = get("spotify_enabled")
    settings_state["spotifySuggestions"] = get("spotify_suggestions")

    settings_state["soundcloudEnabled"] = get("soundcloud_enabled")
    settings_state["soundcloudSuggestions"] = get("soundcloud_suggestions")

    settings_state["jamendoEnabled"] = get("jamendo_enabled")
    settings_state["jamendoSuggestions"] = get("jamendo_suggestions")

    settings_state["backupStream"] = get("backup_stream")

    settings_state["bluetoothScanning"] = redis.get("bluetoothctl_active")
    settings_state["bluetoothDevices"] = redis.get("bluetooth_devices")

    settings_state["feedCava"] = get("feed_cava")
    settings_state["output"] = get("output")

    try:
        with open(os.path.join(conf.BASE_DIR, "config/homewifi")) as f:
            settings_state["homewifiSsid"] = f.read()
    except FileNotFoundError:
        settings_state["homewifiSsid"] = ""

    settings_state["scanProgress"] = redis.get("library_scan_progress")

    try:
        settings_state["homewifiEnabled"] = (
            subprocess.call(["/usr/local/sbin/raveberry/homewifi_enabled"]) != 0
        )
        settings_state["hotspotEnabled"] = (
            subprocess.call(["/usr/local/sbin/raveberry/hotspot_enabled"]) != 0
        )
        settings_state["wifiProtectionEnabled"] = (
            subprocess.call(["/usr/local/sbin/raveberry/wifi_protection_enabled"]) != 0
        )
        settings_state["tunnelingEnabled"] = (
            subprocess.call(["sudo", "/usr/local/sbin/raveberry/tunneling_enabled"])
            != 0
        )
        settings_state["remoteEnabled"] = (
            subprocess.call(["/usr/local/sbin/raveberry/remote_enabled"]) != 0
        )
    except FileNotFoundError:
        settings_state["systemInstall"] = False
    else:
        settings_state["systemInstall"] = True
        with open(os.path.join(conf.BASE_DIR, "config/raveberry.yaml")) as f:
            config = yaml.safe_load(f)
        settings_state["hotspotConfigured"] = config["hotspot"]
        settings_state["remoteConfigured"] = config["remote_key"] is not None

    settings_state["youtubeConfigured"] = redis.get("youtube_available")
    settings_state["spotifyConfigured"] = redis.get("spotify_available")
    settings_state["soundcloudConfigured"] = redis.get("soundcloud_available")
    settings_state["jamendoConfigured"] = redis.get("jamendo_available")

    state["settings"] = settings_state
    return state


@user_passes_test(user_manager.is_admin)
def index(request: WSGIRequest) -> HttpResponse:
    """Renders the /settings page. Only admin is allowed to see this page."""
    from core import urls
    from core.settings import library

    context = base.context(request)
    context["urls"] = urls.settings_paths
    library_path = library.get_library_path()
    if os.path.islink(library_path):
        context["local_library"] = os.readlink(library_path)
    else:
        context["local_library"] = "/"
    context["version"] = conf.VERSION
    return render(request, "settings.html", context)


def update_state() -> None:
    """Sends an update event to all connected clients."""
    send_state(state_dict())



"""This module deals with the "celery" command.
It starts the celery worker and restarts them on file changes."""
import subprocess

from django.core.management.base import BaseCommand
from django.utils import autoreload


def restart_celery() -> None:
    """Kills all celery workers and starts new ones."""
    subprocess.call(["pkill", "-9", "-f", "celery -A core"])

    # https://docs.celeryproject.org/en/stable/userguide/tasks.html
    # use -O fair to be friendly to long-running tasks
    subprocess.call("celery -A core worker -O fair -l INFO".split())


class Command(BaseCommand):
    """Class to register the command"""

    def handle(self, *args, **options) -> None:
        print("Starting celery worker with autoreload...")
        autoreload.run_with_reloader(restart_celery)



# Generated by Django 2.2 on 2021-07-11 17:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("core", "0012_currentsong_last_paused")]

    operations = [
        migrations.AlterField(
            model_name="queuedsong",
            name="external_url",
            field=models.CharField(max_length=2000),
        ),
        migrations.AlterField(
            model_name="queuedsong",
            name="internal_url",
            field=models.CharField(blank=True, max_length=2000, null=True),
        ),
        migrations.AlterField(
            model_name="queuedsong",
            name="stream_url",
            field=models.CharField(blank=True, max_length=2000, null=True),
        ),
        migrations.AlterField(
            model_name="currentsong",
            name="external_url",
            field=models.CharField(max_length=2000),
        ),
        migrations.AlterField(
            model_name="currentsong",
            name="stream_url",
            field=models.CharField(blank=True, max_length=2000, null=True),
        ),
    ]



"""This module provides some utility functions concerning songs."""

import os
import re
from typing import TYPE_CHECKING, Optional

import mutagen.easymp4

import core.settings.storage as storage
from main import settings as conf

if TYPE_CHECKING:
    from typing_extensions import TypedDict
    from core.models import ArchivedPlaylist

    Metadata = TypedDict(  # pylint: disable=invalid-name
        "Metadata",
        {
            "artist": str,
            "title": str,
            "duration": float,
            "internal_url": str,
            "external_url": Optional[str],
            "stream_url": Optional[str],
        },
        total=False,
    )


def get_path(basename: str) -> str:
    """Returns the absolute path for a basename of a file in the cache directory."""
    path = os.path.join(conf.SONGS_CACHE_DIR, basename)
    path = path.replace("~", os.environ["HOME"])
    path = os.path.abspath(path)
    return path


def determine_url_type(url: str) -> str:
    """Returns the service the given url corresponds to."""
    if url.startswith("local_library/"):
        return "local"
    if url.startswith("https://www.youtube.com/"):
        return "youtube"
    if url.startswith("https://open.spotify.com/"):
        return "spotify"
    if url.startswith("https://soundcloud.com/"):
        return "soundcloud"
    if url.startswith("https://www.jamendo.com/"):
        return "jamendo"
    return "unknown"


def determine_playlist_type(archived_playlist: "ArchivedPlaylist") -> str:
    """Uses the url of the first song in the playlist
    to determine the platform where the playlist is from."""
    if archived_playlist.list_id.startswith("playlog"):
        # The playlist was created from play logs and may contain various song types.
        return "playlog"
    first_song = archived_playlist.entries.first()
    if not first_song:
        raise ValueError("Playlist contains no songs.")
    first_song_url = first_song.url
    return determine_url_type(first_song_url)


def format_seconds(seconds: int) -> str:
    """Takes seconds and formats them as [hh:]mm:ss."""

    if seconds < 0:
        return "--:--"

    hours, seconds = seconds // 3600, seconds % 3600
    minutes, seconds = seconds // 60, seconds % 60

    formatted = ""
    if hours > 0:
        formatted += "{:d}:".format(int(hours))
    formatted += "{0:02d}:{1:02d}".format(int(minutes), int(seconds))
    return formatted


def displayname(artist: str, title: str) -> str:
    """Formats the given artist and title as a presentable display name."""
    if artist == "":
        return title
    return artist + " – " + title


def get_metadata(path: str) -> "Metadata":
    """gathers the metadata for the song at the given location.
    'title' and 'duration' is read from tags, the 'url' is built from the location"""

    parsed = mutagen.File(path, easy=True)
    if parsed is None:
        raise ValueError
    metadata: "Metadata" = {}

    if parsed.tags is not None:
        if "artist" in parsed.tags:
            metadata["artist"] = parsed.tags["artist"][0]
        if "title" in parsed.tags:
            metadata["title"] = parsed.tags["title"][0]
    if "artist" not in metadata:
        metadata["artist"] = ""
    if "title" not in metadata:
        metadata["title"] = os.path.split(path)[1]
    if parsed.info is not None and parsed.info.length is not None:
        metadata["duration"] = parsed.info.length
    else:
        metadata["duration"] = -1

    return metadata


def is_forbidden(s: str) -> bool:
    """Returns whether the given string should be filtered according to the forbidden keywords."""
    # We can't access the variable in settings/basic.py
    # since we are in a static context without a reference to bes
    keywords = storage.get("forbidden_keywords")
    words = re.split(r"[,\s]+", keywords.strip())
    # delete empty matches
    words = [word for word in words if word]

    for word in words:
        if re.search(word, s, re.IGNORECASE):
            return True
    return False



"""This module handles all requests concerning the addition of music to the queue."""

from __future__ import annotations

import logging
import os
from typing import Any, Dict, Optional, Union, List, Tuple, cast, Type

from django.conf import settings as conf
from django.core.handlers.wsgi import WSGIRequest
from django.forms.models import model_to_dict
from django.http import HttpResponseBadRequest
from django.http.response import HttpResponse, JsonResponse
from django.shortcuts import render
from django.views.decorators.csrf import csrf_exempt

from django.utils import timezone

import core.musiq.song_utils as song_utils
import core.settings.storage as storage
from core import util, base, redis, user_manager
from core.models import CurrentSong
from core.models import QueuedSong
from core.musiq.localdrive import LocalSongProvider
from core.musiq.music_provider import MusicProvider, WrongUrlError, ProviderError
from core.musiq.song_provider import SongProvider
from core.musiq.playlist_provider import PlaylistProvider
from core.state_handler import send_state

queue = QueuedSong.objects


def start() -> None:
    import core.musiq.controller as controller
    import core.musiq.playback as playback

    controller.start()
    playback.start()


def get_alarm_metadata() -> "Metadata":
    return {
        "artist": "Raveberry",
        "title": "ALARM!",
        "duration": song_utils.get_metadata(
            os.path.join(conf.BASE_DIR, "config/sounds/alarm.m4a")
        )["duration"],
        "internal_url": "alarm",
        "external_url": "https://raveberry.party/alarm",
        "stream_url": None,
    }


def do_request_music(
    request_ip: str,
    query: str,
    key: Optional[int],
    playlist: bool,
    platform: str,
    archive: bool = True,
    manually_requested: bool = True,
) -> Tuple[bool, str, Optional[int]]:
    """Performs the actual requesting of the music, not an endpoint.
    Enqueues the requested song or playlist into the queue, using appropriate providers.
    Returns a 3-tuple: successful, message, queue_key"""
    providers: List[MusicProvider] = []

    provider: MusicProvider
    music_provider_class: Union[Type[PlaylistProvider], Type[SongProvider]]
    local_provider_class: Type[MusicProvider]
    jamendo_provider_class: Type[MusicProvider]
    soundcloud_provider_class: Type[MusicProvider]
    spotify_provider_class: Type[MusicProvider]
    youtube_provider_class: Type[MusicProvider]

    if playlist:
        music_provider_class = PlaylistProvider
        local_provider_class = PlaylistProvider
        if storage.get("jamendo_enabled"):
            from core.musiq.jamendo import JamendoPlaylistProvider

            jamendo_provider_class = JamendoPlaylistProvider
        if storage.get("soundcloud_enabled"):
            from core.musiq.soundcloud import SoundcloudPlaylistProvider

            soundcloud_provider_class = SoundcloudPlaylistProvider
        if storage.get("spotify_enabled"):
            from core.musiq.spotify import SpotifyPlaylistProvider

            spotify_provider_class = SpotifyPlaylistProvider
        if storage.get("youtube_enabled"):
            from core.musiq.youtube import YoutubePlaylistProvider

            youtube_provider_class = YoutubePlaylistProvider
    else:
        music_provider_class = SongProvider
        local_provider_class = LocalSongProvider
        if storage.get("jamendo_enabled"):
            from core.musiq.jamendo import JamendoSongProvider

            jamendo_provider_class = JamendoSongProvider
        if storage.get("soundcloud_enabled"):
            from core.musiq.soundcloud import SoundcloudSongProvider

            soundcloud_provider_class = SoundcloudSongProvider
        if storage.get("spotify_enabled"):
            from core.musiq.spotify import SpotifySongProvider

            spotify_provider_class = SpotifySongProvider
        if storage.get("youtube_enabled"):
            from core.musiq.youtube import YoutubeSongProvider

            youtube_provider_class = YoutubeSongProvider

    if key is not None:
        # an archived entry was requested.
        # The key determines the Provider
        provider = music_provider_class.create(query, key)
        if provider is None:
            return False, "No provider found for requested key", None
        providers.append(provider)
    else:
        if platform == "local":
            # local music can only be searched explicitly
            providers.append(local_provider_class(query, key))
        if storage.get("soundcloud_enabled"):
            try:
                soundcloud_provider = soundcloud_provider_class(query, key)
                if platform == "soundcloud":
                    providers.insert(0, soundcloud_provider)
                else:
                    providers.append(soundcloud_provider)
            except WrongUrlError:
                pass
        if storage.get("spotify_enabled"):
            try:
                spotify_provider = spotify_provider_class(query, key)
                if platform == "spotify":
                    providers.insert(0, spotify_provider)
                else:
                    providers.append(spotify_provider)
            except WrongUrlError:
                pass
        if storage.get("jamendo_enabled"):
            try:
                jamendo_provider = jamendo_provider_class(query, key)
                if platform == "jamendo":
                    providers.insert(0, jamendo_provider)
                else:
                    providers.append(jamendo_provider)
            except WrongUrlError:
                pass
        if storage.get("youtube_enabled"):
            try:
                youtube_provider = youtube_provider_class(query, key)
                if platform == "youtube":
                    providers.insert(0, youtube_provider)
                else:
                    providers.append(youtube_provider)
            except WrongUrlError:
                pass

    if not providers:
        return False, "No backend configured to handle your request.", None

    fallback = False
    for i, provider in enumerate(providers):
        try:
            provider.request(
                request_ip, archive=archive, manually_requested=manually_requested
            )
            # the current provider could provide the song, don't try the other ones
            break
        except ProviderError:
            # this provider cannot provide this song, use the next provider
            # if this was the last provider, show its error
            # in new music only mode, do not allow fallbacks
            if storage.get("new_music_only") or i == len(providers) - 1:
                return False, provider.error, None
            fallback = True
    message = provider.ok_message
    queue_key = None
    if not playlist:
        queued_song = cast(SongProvider, provider).queued_song
        if not queued_song:
            logging.error(
                "no placeholder was created for query '%s' and key '%s'", query, key
            )
            return False, "No placeholder was created", None
        queue_key = queued_song.id
    if fallback:
        message += " (used fallback)"
    return True, message, queue_key


# accessed by the discord bot
@csrf_exempt
def request_music(request: WSGIRequest) -> HttpResponse:
    """Endpoint to request music. Calls internal function."""
    key = request.POST.get("key")
    query = request.POST.get("query")
    playlist = request.POST.get("playlist") == "true"
    platform = request.POST.get("platform")

    if query is None or not platform:
        return HttpResponseBadRequest(
            "query, playlist and platform have to be specified."
        )
    ikey = None
    if key:
        ikey = int(key)

    # only get ip on user requests
    request_ip = user_manager.get_client_ip(request)

    successful, message, queue_key = do_request_music(
        request_ip, query, ikey, playlist, platform
    )
    if not successful:
        return HttpResponseBadRequest(message)
    return JsonResponse({"message": message, "key": queue_key})


def request_radio(request: WSGIRequest) -> HttpResponse:
    """Endpoint to request radio for the current song."""
    # only get ip on user requests
    request_ip = user_manager.get_client_ip(request)

    try:
        current_song = CurrentSong.objects.get()
    except CurrentSong.DoesNotExist:
        return HttpResponseBadRequest("Need a song to play the radio")
    provider = SongProvider.create(external_url=current_song.external_url)
    return provider.request_radio(request_ip)


def index(request: WSGIRequest) -> HttpResponse:
    """Renders the /musiq page."""
    from core import urls

    context = base.context(request)
    context["urls"] = urls.musiq_paths
    context["additional_keywords"] = storage.get("additional_keywords")
    context["forbidden_keywords"] = storage.get("forbidden_keywords")
    context["embed_stream"] = storage.get("embed_stream")
    context["dynamic_embedded_stream"] = storage.get("dynamic_embedded_stream")
    return render(request, "musiq.html", context)


def state_dict() -> Dict[str, Any]:
    state = base.state_dict()

    musiq_state = {}

    musiq_state["paused"] = storage.get("paused")
    musiq_state["shuffle"] = storage.get("shuffle")
    musiq_state["repeat"] = storage.get("repeat")
    musiq_state["autoplay"] = storage.get("autoplay")
    musiq_state["volume"] = storage.get("volume")

    try:
        current_song = CurrentSong.objects.get()
        current_song_dict = model_to_dict(current_song)
        current_song_dict = util.camelize(current_song_dict)
        current_song_dict["durationFormatted"] = song_utils.format_seconds(
            current_song_dict["duration"]
        )
        musiq_state["currentSong"] = current_song_dict

        paused = storage.get("paused")
        if paused:
            progress = (current_song.last_paused - current_song.created).total_seconds()
        else:
            progress = (timezone.now() - current_song.created).total_seconds()
        progress /= current_song.duration
        musiq_state["progress"] = progress * 100
    except CurrentSong.DoesNotExist:
        musiq_state["currentSong"] = None
        musiq_state["paused"] = True
        musiq_state["progress"] = 0

    song_queue = []
    total_time = 0
    all_songs = queue.all()
    if storage.get("voting_enabled"):
        all_songs = all_songs.order_by("-votes", "index")
    for song in all_songs:
        song_dict = model_to_dict(song)
        song_dict = util.camelize(song_dict)
        song_dict["durationFormatted"] = song_utils.format_seconds(
            song_dict["duration"]
        )
        song_queue.append(song_dict)
        if song_dict["duration"] < 0:
            # skip duration of placeholders
            continue
        total_time += song_dict["duration"]
    musiq_state["totalTimeFormatted"] = song_utils.format_seconds(total_time)
    musiq_state["songQueue"] = song_queue

    if state["alarm"]:
        musiq_state["currentSong"] = {
            "queueKey": -1,
            "manuallyRequested": False,
            "votes": 0,
            "created": "",
            **util.camelize(get_alarm_metadata()),
        }
        musiq_state["progress"] = 0
        musiq_state["paused"] = False
    elif redis.get("backup_playing"):
        musiq_state["currentSong"] = {
            "queueKey": -1,
            "manuallyRequested": False,
            "votes": 0,
            "internalUrl": "backup_stream",
            "externalUrl": storage.get("backup_stream"),
            "artist": "",
            "title": "Backup Stream",
            "duration": 60 * 60 * 24,
            "created": "",
        }
        musiq_state["paused"] = False

    state["musiq"] = musiq_state
    return state


def update_state() -> None:
    """Sends an update event to all connected clients."""
    send_state(state_dict())



"""Contains the core app configuration."""
import sys
import logging
import os
import atexit

from django.apps import AppConfig
from django.conf import settings as conf
from watson import search as watson


class CoreConfig(AppConfig):
    """This is the configuration for the core app."""

    name = "core"

    def ready(self) -> None:
        if "celery" in sys.argv and os.environ.get("RUN_MAIN", None) == "true":
            # if the development celery process starts,
            # have it import all modules containing celery tasks
            # this way, its autoreload is notified on changes in these modules

            for module in conf.CELERY_IMPORTS:
                __import__(module)

            # these are used by the lights worker but imported locally.
            # in order to trigger autoreloads they need to imported here as well
            import core.lights.ring
            import core.lights.wled
            import core.lights.strip
            import core.lights.screen

            return

        # ready is called for every management command and for autoreload
        # only start raveberry when
        # in debug mode and the main application is run (not autoreload)
        # or in prod mode (run by daphne)
        start_raveberry = (
            "runserver" in sys.argv and os.environ.get("RUN_MAIN", None) == "true"
        ) or (sys.argv[0].endswith("daphne"))

        if (
            start_raveberry
            or conf.TESTING
            or any("watson" in arg for arg in sys.argv)
            or any("celery" in arg for arg in sys.argv)
        ):
            watson.register(
                self.get_model("ArchivedSong"),
                fields=("url", "artist", "title", "queries__query"),
                store=("id", "title", "url", "artist"),
            )
            watson.register(
                self.get_model("ArchivedPlaylist"),
                fields=("title", "queries__query"),
                store=("id", "title", "counter"),
            )

        if conf.TESTING:
            import core.musiq.controller as controller

            # when MopidyAPI instances are created too often,
            # mopidy runs into a "Set changed size during iteration" error.
            # Thus we initialize the interface once and not for every testcase
            controller.start()

        if start_raveberry:
            import core.celery as celery
            import core.redis as redis
            import core.musiq.musiq as musiq
            import core.musiq.playback as playback
            import core.settings.basic as basic
            import core.settings.platforms as platforms
            import core.lights.worker as worker

            logging.info("starting raveberry")

            redis.start()
            celery.start()

            worker.start()
            musiq.start()
            basic.start()
            platforms.start()

            def stop_workers() -> None:
                # wake up the playback thread and stop it
                redis.set("stop_playback_loop", True)
                playback.queue_changed.set()

                # wake the buzzer thread so it exits
                playback.buzzer_stopped.set()

                # wake up the listener thread with an instruction to stop the lights worker
                redis.publish("lights_settings_changed", "stop")

            atexit.register(stop_workers)



from django.core.management.base import BaseCommand, CommandError
from django.template.loader import get_template
from django.template.loader import TemplateDoesNotExist
from bs4 import BeautifulSoup
import json


class Command(BaseCommand):
    help = "Renders the specified template, splits it into head and body and writes it to the given output files"

    def add_arguments(self, parser):
        parser.add_argument("template")
        parser.add_argument("head_file")
        parser.add_argument("body_file")
        parser.add_argument("context", nargs="?")

    def handle(self, *args, **options):
        try:
            template = get_template(options["template"])
        except TemplateDoesNotExist:
            raise CommandError('Template "%s" does not exist' % options["template"])

        if options["context"]:
            context = json.loads(options["context"])
        else:
            context = {}
        html = template.render(context)

        soup = BeautifulSoup(html, "html.parser")
        with open(options["head_file"], "w") as f:
            f.write(soup.head.prettify())
        with open(options["body_file"], "w") as f:
            f.write(soup.body.prettify())



"""This module handles all controls that influence visualization in general."""

from __future__ import annotations

import socket
from functools import wraps
from typing import cast, Tuple, Callable

from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponseBadRequest, HttpResponse, HttpResponseForbidden

from core import user_manager, redis
from core.lights import lights, worker
from core.settings import storage


def control(func: Callable) -> Callable:
    """A decorator for functions that control the lights.
    Every control changes the views state and returns an empty response.
    Mod privileges are required to control the lights."""

    def _decorator(request: WSGIRequest) -> HttpResponse:
        if not user_manager.has_controls(request.user):
            return HttpResponseForbidden()

        response = func(request)
        lights.update_state()
        if response is not None:
            return response
        return HttpResponse()

    return wraps(func)(_decorator)


def _notify_settings_changed(settings: str) -> None:
    # notifies the worker thread that the given settings (eg <device> or "base") changed
    redis.publish(f"lights_settings_changed", settings)


def alarm_started() -> None:
    """Notifies the lights worker that the alarm was started."""
    _notify_settings_changed("alarm_started")


def alarm_stopped() -> None:
    """Notifies the lights worker that the alarm was stopped."""
    _notify_settings_changed("alarm_stopped")


def set_program(device: str, program: str) -> None:
    """Updates the given device to the given program."""
    current_program = storage.get(f"{device}_program")
    storage.set(f"last_{device}_program", current_program)
    storage.set(f"{device}_program", program)
    _notify_settings_changed(device)


@control
def set_lights_shortcut(request: WSGIRequest) -> None:
    """Stores the current lights state and restores the previous one."""
    should_enable = request.POST.get("value") == "true"
    is_enabled = (
        storage.get("ring_program") != "Disabled"
        or storage.get("wled_program") != "Disabled"
        or storage.get("strip_program") != "Disabled"
    )
    if should_enable == is_enabled:
        return
    if should_enable:
        for device in ["ring", "wled", "strip"]:
            set_program(device, storage.get(f"last_{device}_program"))
    else:
        for device in ["ring", "wled", "strip"]:
            set_program(device, "Disabled")


@control
def set_program_speed(request: WSGIRequest) -> None:
    """Updates the global speed of programs supporting it."""
    value = float(request.POST.get("value"))  # type: ignore
    storage.set("program_speed", value)
    _notify_settings_changed("base")


@control
def set_fixed_color(request: WSGIRequest) -> None:
    """Updates the static color used for some programs."""
    hex_col = request.POST.get("value").lstrip("#")  # type: ignore
    # raises IndexError on wrong input, caught in option decorator
    color = tuple(int(hex_col[i : i + 2], 16) / 255 for i in (0, 2, 4))
    # https://github.com/python/mypy/issues/5068
    color = cast(Tuple[float, float, float], color)
    storage.set("fixed_color", str(color))
    _notify_settings_changed("base")


def _handle_program_request(device: str, request: WSGIRequest) -> None:
    program = request.POST.get("value")
    if not program:
        return
    if program == storage.get(f"{device}_program"):
        # the program doesn't change, return immediately
        return
    set_program(device, program)


def _handle_brightness_request(device: str, request: WSGIRequest) -> None:
    # raises ValueError on wrong input, caught in option decorator
    value = float(request.POST.get("value"))  # type: ignore
    storage.set(f"{device}_brightness", value)
    _notify_settings_changed(device)


def _handle_monochrome_request(device: str, request: WSGIRequest) -> None:
    # raises ValueError on wrong input, caught in option decorator
    enabled = request.POST.get("value") == "true"  # type: ignore
    storage.set(f"{device}_monochrome", enabled)
    _notify_settings_changed(device)


@control
def set_ring_program(request: WSGIRequest) -> None:
    """Updates the ring program."""
    _handle_program_request("ring", request)


@control
def set_ring_brightness(request: WSGIRequest) -> None:
    """Updates the ring brightness."""
    _handle_brightness_request("ring", request)


@control
def set_ring_monochrome(request: WSGIRequest) -> None:
    """Sets whether the ring should be in one color only."""
    _handle_monochrome_request("ring", request)


@control
def set_wled_led_count(request: WSGIRequest) -> None:
    """Updates the wled led_count."""
    value = int(request.POST.get("value"))  # type: ignore
    if not (2 <= value <= 490):
        return
    storage.set("wled_led_count", value)
    _notify_settings_changed("wled")


@control
def set_wled_ip(request: WSGIRequest) -> None:
    """Updates the wled ip."""
    value = request.POST.get("value")  # type: ignore
    try:
        socket.inet_aton(value)
    except socket.error:
        return
    storage.set("wled_ip", value)
    _notify_settings_changed("wled")


@control
def set_wled_port(request: WSGIRequest) -> None:
    """Updates the wled port."""
    value = int(request.POST.get("value"))  # type: ignore
    if not (1024 <= value <= 65535):
        return
    storage.set("wled_port", value)
    _notify_settings_changed("wled")


@control
def set_wled_program(request: WSGIRequest) -> None:
    """Updates the wled program."""
    _handle_program_request("wled", request)


@control
def set_wled_brightness(request: WSGIRequest) -> None:
    """Updates the wled brightness."""
    _handle_brightness_request("wled", request)


@control
def set_wled_monochrome(request: WSGIRequest) -> None:
    """Sets whether the wled should be in one color only."""
    _handle_monochrome_request("wled", request)


@control
def set_strip_program(request: WSGIRequest) -> None:
    """Updates the strip program."""
    _handle_program_request("strip", request)


@control
def set_strip_brightness(request: WSGIRequest) -> None:
    """Updates the strip brightness."""
    _handle_brightness_request("strip", request)


@control
def set_screen_program(request: WSGIRequest) -> None:
    """Updates the screen program."""
    _handle_program_request("screen", request)


@control
def adjust_screen(_request: WSGIRequest) -> HttpResponse:
    """Adjusts the resolution of the screen."""
    if storage.get("screen_program") != "Disabled":
        return HttpResponseBadRequest("Disable the screen program before readjusting")
    # _notify_settings_changed("adjust_screen")
    # After disabling the screen program, it cannot be started again
    # (pi3d only shows the black background)
    # We restart the lights worker so the library is reloaded,
    # allowing the screen program to be shown again.
    # During initialization, the screen will adjust itself.
    redis.publish("lights_settings_changed", "stop")
    worker.start()

    return HttpResponse()



"""This module handles all settings regarding the music platforms."""
from __future__ import annotations

import importlib
import os
import subprocess

from django.conf import settings as conf
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse
from django.http import HttpResponseBadRequest

from core import redis
from core.settings import library, system
from core.settings.settings import control
from core.settings import storage


def start() -> None:
    """Initializes this module by checking which platforms are available to use."""

    # local songs are enabled if a library is set
    local_enabled = os.path.islink(library.get_library_path())
    storage.set("local_enabled", local_enabled)

    # in the docker container all dependencies are installed
    youtube_available = (
        conf.DOCKER or importlib.util.find_spec("youtube_dl") is not None
    )
    redis.set("youtube_available", youtube_available)
    if not youtube_available:
        # if youtube is not available, overwrite the database to disable it
        storage.set("youtube_enabled", False)

    # Spotify has no python dependencies we could easily check.
    try:
        spotify_available = (
            conf.DOCKER
            or "[spotify]"
            in subprocess.check_output(["mopidy", "config"], stderr=subprocess.DEVNULL)
            .decode()
            .splitlines()
        )
    except FileNotFoundError:
        # mopidy is not installed (eg in docker). Since we can't check, enable
        spotify_available = True
    redis.set("spotify_available", spotify_available)
    if not spotify_available:
        storage.set("spotify_enabled", False)

    soundcloud_available = (
        conf.DOCKER or importlib.util.find_spec("soundcloud") is not None
    )
    redis.set("soundcloud_available", soundcloud_available)
    if not soundcloud_available:
        storage.set("soundcloud_enabled", False)

    # Jamendo has no python dependencies we could easily check.
    try:
        jamendo_available = (
            conf.DOCKER
            or "[jamendo]"
            in subprocess.check_output(["mopidy", "config"], stderr=subprocess.DEVNULL)
            .decode()
            .splitlines()
        )
    except FileNotFoundError:
        jamendo_available = True
    redis.set("jamendo_available", jamendo_available)
    if not jamendo_available:
        storage.set("jamendo_enabled", False)


@control
def set_youtube_enabled(request: WSGIRequest):
    """Enables or disables youtube to be used as a song provider."""
    enabled = request.POST.get("value") == "true"
    storage.set("youtube_enabled", enabled)


@control
def set_youtube_suggestions(request: WSGIRequest):
    """Sets the number of online suggestions from youtube to be shown."""
    value = int(request.POST.get("value"))  # type: ignore
    storage.set("youtube_suggestions", value)


def _set_extension_enabled(extension, enabled) -> HttpResponse:
    if enabled:
        if conf.DOCKER:
            response = HttpResponse(
                "Make sure you provided mopidy with correct credentials."
            )
        else:
            extensions = system.check_mopidy_extensions()
            functional, message = extensions[extension]
            if not functional:
                return HttpResponseBadRequest(message)
            response = HttpResponse(message)
    else:
        response = HttpResponse("Disabled extension")
    storage.set(f"{extension}_enabled", enabled)
    return response


@control
def set_spotify_enabled(request: WSGIRequest) -> HttpResponse:
    """Enables or disables spotify to be used as a song provider.
    Makes sure mopidy has correct spotify configuration."""
    enabled = request.POST.get("value") == "true"
    return _set_extension_enabled("spotify", enabled)


@control
def set_spotify_suggestions(request: WSGIRequest):
    """Sets the number of online suggestions from spotify to be shown."""
    value = int(request.POST.get("value"))  # type: ignore
    storage.set("spotify_suggestions", value)


@control
def set_spotify_credentials(request: WSGIRequest) -> HttpResponse:
    """Update spotify credentials."""
    username = request.POST.get("username")
    password = request.POST.get("password")
    client_id = request.POST.get("client_id")
    client_secret = request.POST.get("client_secret")

    if not username or not password or not client_id or not client_secret:
        return HttpResponseBadRequest("All fields are required")

    storage.set("spotify_username", username)
    storage.set("spotify_password", password)
    storage.set("spotify_client_id", client_id)
    storage.set("spotify_client_secret", client_secret)

    system.update_mopidy_config("pulse")
    return HttpResponse("Updated credentials")


@control
def set_soundcloud_enabled(request: WSGIRequest) -> HttpResponse:
    """Enables or disables soundcloud to be used as a song provider.
    Makes sure mopidy has correct soundcloud configuration."""
    enabled = request.POST.get("value") == "true"
    return _set_extension_enabled("soundcloud", enabled)


@control
def set_soundcloud_suggestions(request: WSGIRequest):
    """Sets the number of online suggestions from soundcloud to be shown."""
    value = int(request.POST.get("value"))  # type: ignore
    storage.set("soundcloud_suggestions", value)


@control
def set_soundcloud_credentials(request: WSGIRequest) -> HttpResponse:
    """Update soundcloud credentials."""
    auth_token = request.POST.get("auth_token")

    if not auth_token:
        return HttpResponseBadRequest("All fields are required")

    storage.set("soundcloud_auth_token", auth_token)

    system.update_mopidy_config("pulse")
    return HttpResponse("Updated credentials")


@control
def set_jamendo_enabled(request: WSGIRequest) -> HttpResponse:
    """Enables or disables jamendo to be used as a song provider.
    Makes sure mopidy has correct jamendo configuration."""
    enabled = request.POST.get("value") == "true"
    return _set_extension_enabled("jamendo", enabled)


@control
def set_jamendo_suggestions(request: WSGIRequest):
    """Sets the number of online suggestions from jamendo to be shown."""
    value = int(request.POST.get("value"))  # type: ignore
    storage.set("jamendo_suggestions", value)


@control
def set_jamendo_credentials(request: WSGIRequest) -> HttpResponse:
    """Update jamendo credentials."""
    client_id = request.POST.get("client_id")

    if not client_id:
        return HttpResponseBadRequest("All fields are required")

    storage.set("jamendo_client_id", client_id)

    system.update_mopidy_config("pulse")
    return HttpResponse("Updated credentials")



"""This module contains all programs used for visualization (except complex screen programs)."""

from __future__ import annotations

import colorsys
import errno
import logging
import math
import os
import subprocess
from typing import Tuple, List, cast, Optional, TYPE_CHECKING

from django.conf import settings

if TYPE_CHECKING:
    from core.lights.worker import DeviceManager


def stretched_hues(led_count: int, offset: float = 0) -> List[float]:
    # map the leds to rainbow colors from red over green to blue
    # (without pink-> hue values in [0, ⅔]
    # stretch the outer regions (red and blue) and compress the inner region (green)
    return [
        (2 / 3)
        * 1
        / (1 + math.e ** (-6 * math.e * (((offset + led / led_count) % 1) - 0.5)))
        for led in range(0, led_count)
    ]


class VizProgram:
    """The base class for all programs."""

    def __init__(self, manager: "DeviceManager") -> None:
        self.manager = manager
        self.consumers = 0
        self.name = "Unknown"

    def start(self) -> None:
        """Initializes the program, allocates resources."""

    def use(self) -> None:
        """Tells the program that it is used by another consumer.
        Starts the program if this is the first usage."""
        if self.consumers == 0:
            self.start()
        self.consumers += 1

    def stop(self) -> None:
        """Stops the program, releases resources."""

    def release(self) -> None:
        """Tells the program that one consumer does not use it anymore.
        Stops the program if this was the last one."""
        self.consumers -= 1
        if self.consumers == 0:
            self.stop()


class ScreenProgram(VizProgram):
    """The base class for all screen visualization programs."""

    def draw(self) -> None:
        """Called every frame. Updates the screen."""
        raise NotImplementedError()

    def increase_resolution(self) -> None:
        """Called if there is time in the loop to spare.
        Increases the system load, but also increases quality."""

    def decrease_resolution(self) -> None:
        """Called if rendering takes too long.
        Decreases the quality and speeds up the draw call."""


class LedProgram(VizProgram):
    """The base class for all led visualization programs."""

    def compute(self) -> None:
        """Is called once per update. Computation should happen here,
        so they can be reused in the returning functions"""

    def ring_colors(self) -> List[Tuple[float, float, float]]:
        """Returns the colors for the ring, one rgb tuple for each led."""
        raise NotImplementedError()

    def wled_colors(self) -> List[Tuple[float, float, float]]:
        """Returns the colors for WLED, one rgb tuple for each led."""
        raise NotImplementedError()

    def strip_color(self) -> Tuple[float, float, float]:
        """Returns the rgb values for the strip."""
        raise NotImplementedError()


class Disabled(LedProgram, ScreenProgram):
    """A null class to represent inactivity."""

    def __init__(self, manager: "DeviceManager") -> None:
        super().__init__(manager)
        self.name = "Disabled"

    def draw(self) -> None:
        raise NotImplementedError()

    def ring_colors(self) -> List[Tuple[float, float, float]]:
        raise NotImplementedError()

    def wled_colors(self) -> List[Tuple[float, float, float]]:
        raise NotImplementedError()

    def strip_color(self) -> Tuple[float, float, float]:
        raise NotImplementedError()


class Fixed(LedProgram):
    """Show one fixed color only. The color is controlled in the "DeviceManager" class."""

    def __init__(self, manager: "DeviceManager") -> None:
        super().__init__(manager)
        self.manager = manager
        self.name = "Fixed"

    def compute(self) -> None:
        # show a red color if the alarm is active
        alarm_factor = self.manager.alarm_program.factor
        if alarm_factor != -1.0:
            self.manager.fixed_color = (alarm_factor, 0, 0)

    def ring_colors(self) -> List[Tuple[float, float, float]]:
        return [self.manager.fixed_color for _ in range(self.manager.ring.LED_COUNT)]

    def wled_colors(self) -> List[Tuple[float, float, float]]:
        return [self.manager.fixed_color for _ in range(self.manager.wled.led_count)]

    def strip_color(self) -> Tuple[float, float, float]:
        return self.manager.fixed_color


class Rainbow(LedProgram):
    """Continuously cycles through all colors. Affected by the speed setting."""

    def __init__(self, manager: "DeviceManager") -> None:
        super().__init__(manager)
        self.manager = manager
        self.name = "Rainbow"
        self.program_duration = 1
        self.time_passed = 0.0
        self.current_fraction = 0.0

    def start(self) -> None:
        self.time_passed = 0.0

    def compute(self) -> None:
        self.time_passed += self.manager.seconds_per_frame * self.manager.program_speed
        self.time_passed %= self.program_duration
        self.current_fraction = self.time_passed / self.program_duration

    def _colors(self, led_count) -> List[Tuple[float, float, float]]:
        return [
            colorsys.hsv_to_rgb(hue, 1, 1)
            for hue in stretched_hues(led_count, self.current_fraction)
        ]

    def ring_colors(self) -> List[Tuple[float, float, float]]:
        return self._colors(self.manager.ring.LED_COUNT)

    def wled_colors(self) -> List[Tuple[float, float, float]]:
        return self._colors(self.manager.wled.led_count)

    def strip_color(self) -> Tuple[float, float, float]:
        return colorsys.hsv_to_rgb(self.current_fraction, 1, 1)


class Adaptive(LedProgram):
    """Dynamically reacts to the currently played music.
    Low frequencies are represented by red, high ones by blue."""

    def __init__(self, manager: "DeviceManager") -> None:
        super().__init__(manager)
        self.manager = manager
        self.name = "Rave"
        self.cava = self.manager.cava_program

        # RING
        ring_hues = stretched_hues(self.manager.ring.LED_COUNT)
        self.ring_base_colors = [colorsys.hsv_to_rgb(hue, 1, 1) for hue in ring_hues]

        # WLED
        # identical to ring, but with a different number of leds
        wled_hues = stretched_hues(self.manager.wled.led_count)
        self.wled_base_colors = [colorsys.hsv_to_rgb(hue, 1, 1) for hue in wled_hues]

        # STRIP
        # distribute frequencies over the three leds. Don't use hard cuts, but smooth functions
        # the functions add up to one at every point and each functions integral is a third
        self.strip_granularity = 16
        self.red_coeffs = [
            -1
            / (
                1
                + math.e ** (-6 * math.e * (led / (self.strip_granularity - 1) - 1 / 3))
            )
            + 1
            for led in range(0, self.strip_granularity)
        ]
        self.blue_coeffs = [
            1
            / (
                1
                + math.e ** (-6 * math.e * (led / (self.strip_granularity - 1) - 2 / 3))
            )
            for led in range(0, self.strip_granularity)
        ]
        self.green_coeffs = [
            1 - self.red_coeffs[led] - self.blue_coeffs[led]
            for led in range(0, self.strip_granularity)
        ]

    def start(self) -> None:
        self.cava.use()

    def compute(self) -> None:
        pass

    def _aggregate_frame(self, led_count) -> List[float]:
        # aggregate the length of cavas frame into a list the length of the number of leds we have.
        # This reduces computation time.
        values_per_led = len(self.cava.current_frame) // led_count
        left = len(self.cava.current_frame) - values_per_led * led_count

        start = 0

        aggregated = []
        for _ in range(led_count):
            end = start + values_per_led
            bin_size = values_per_led
            if left > 0:
                end += 1
                left -= 1
                bin_size += 1

            aggregated.append(sum(self.cava.current_frame[start:end]) / bin_size)
            start += values_per_led
        return aggregated

    def ring_colors(self) -> List[Tuple[float, float, float]]:
        aggregated = self._aggregate_frame(self.manager.ring.LED_COUNT)
        colors = [
            tuple(factor * val for val in color)
            for factor, color in zip(aggregated, self.ring_base_colors)
        ]
        # https://github.com/python/mypy/issues/5068
        return cast(List[Tuple[float, float, float]], colors)

    def wled_colors(self) -> List[Tuple[float, float, float]]:
        aggregated = self._aggregate_frame(self.manager.wled.led_count)
        colors = [
            tuple(factor * val for val in color)
            for factor, color in zip(aggregated, self.wled_base_colors)
        ]
        # https://github.com/python/mypy/issues/5068
        return cast(List[Tuple[float, float, float]], colors)

    def strip_color(self) -> Tuple[float, float, float]:
        aggregated = self._aggregate_frame(self.strip_granularity)
        red = (
            sum(coeff * val for coeff, val in zip(self.red_coeffs, aggregated))
            * 3
            / self.strip_granularity
        )
        green = (
            sum(coeff * val for coeff, val in zip(self.green_coeffs, aggregated))
            * 3
            / self.strip_granularity
        )
        blue = (
            sum(coeff * val for coeff, val in zip(self.blue_coeffs, aggregated))
            * 3
            / self.strip_granularity
        )
        red = min(1.0, red)
        green = min(1.0, green)
        blue = min(1.0, blue)
        return red, green, blue

    def stop(self) -> None:
        self.cava.release()


class Alarm(VizProgram):
    """This program makes the leds flash red in sync to the played sound.
    Only computes the brightness, does not display it."""

    def __init__(self, manager: "DeviceManager") -> None:
        super().__init__(manager)
        self.manager = manager
        self.name = "Alarm"
        self.time_passed = 0.0
        self.sound_count = 0
        self.increasing_duration = 0.45
        self.decreasing_duration = 0.8
        self.sound_duration = 2.1
        self.sound_repetition = 2.5
        self.factor = -1.0

    def start(self) -> None:
        self.time_passed = 0.0
        self.sound_count = 0
        self.factor = 0

    def compute(self) -> None:
        """If active, compute the brightness for the red color,
        depending on the time that has passed since starting the sound."""
        # do not compute if the alarm is not active
        if self.consumers == 0:
            return
        self.time_passed += self.manager.seconds_per_frame
        if self.time_passed >= self.sound_repetition:
            self.sound_count += 1
            self.time_passed %= self.sound_repetition

        if self.sound_count >= 4:
            self.factor = 0
            return
        if self.time_passed < self.increasing_duration:
            self.factor = self.time_passed / self.increasing_duration
        elif self.time_passed < self.sound_duration - self.decreasing_duration:
            self.factor = 1
        elif self.time_passed < self.sound_duration:
            self.factor = (
                1
                - (self.time_passed - (self.sound_duration - self.decreasing_duration))
                / self.decreasing_duration
            )
        else:
            self.factor = 0

    def stop(self) -> None:
        self.factor = -1.0


class Cava(VizProgram):
    """This Program manages the interaction with cava.
    It provides the current frequencies for other programs to use."""

    def __init__(self, manager: "DeviceManager") -> None:
        super().__init__(manager)
        self.manager = manager

        self.cava_fifo_path = os.path.join(settings.BASE_DIR, "config/cava_fifo")

        # Keep these configurations in sync with config/cava.config
        self.bars = 256
        self.bit_format = 8

        self.frame_length = self.bars * (self.bit_format // 8)

        self.current_frame: List[float] = []
        self.growing_frame = b""
        self.cava_process: Optional[subprocess.Popen[bytes]] = None
        self.cava_fifo = -1

    def start(self) -> None:
        self.current_frame = [0 for _ in range(self.bars)]
        self.growing_frame = b""
        try:
            # delete old contents of the pipe
            os.remove(self.cava_fifo_path)
        except FileNotFoundError:
            # the file does not exist
            pass
        try:
            os.mkfifo(self.cava_fifo_path)
        except FileExistsError:
            # the file already exists
            logging.info("%s already exists while starting", self.cava_fifo_path)

        self.cava_process = subprocess.Popen(
            ["cava", "-p", os.path.join(settings.BASE_DIR, "config/cava.config")],
            cwd=settings.BASE_DIR,
            env={"PULSE_SERVER": "127.0.0.1", **os.environ},
        )
        # cava_fifo = open(cava_fifo_path, 'r')
        self.cava_fifo = os.open(self.cava_fifo_path, os.O_RDONLY | os.O_NONBLOCK)

    def compute(self) -> None:
        """If active, read output from the cava program.
        Make sure that the most recent frame is always fully available,
        Stores incomplete frames for the next update."""
        # do not compute if no program uses cava
        if self.consumers == 0:
            return
        # read the fifo until we get to the current frame
        while True:
            try:
                read = os.read(
                    self.cava_fifo, self.frame_length - len(self.growing_frame)
                )
                if read == b"":
                    return
                self.growing_frame += read
            except OSError as e:
                if e.errno == errno.EAGAIN or e.errno == errno.EWOULDBLOCK:
                    # there were not enough bytes for a whole frame, keep the old frame
                    return

            # we read a whole frame, update the factors
            if len(self.growing_frame) == self.frame_length:
                # vol = max(0.01, get("volume"))
                # self.current_frame = [int(b) / 255 / vol for b in self.growing_frame]
                self.current_frame = [int(b) / 255 for b in self.growing_frame]
                self.growing_frame = b""

    def stop(self) -> None:
        try:
            os.close(self.cava_fifo)
        except OSError as e:
            logging.info("fifo already closed: %s", e)
        except TypeError as e:
            logging.info("fifo does not exist: %s", e)

        if self.cava_process:
            self.cava_process.terminate()

        try:
            os.remove(self.cava_fifo_path)
        except FileNotFoundError as e:
            # the file was already deleted
            logging.info("%s not found while deleting: %s", self.cava_fifo_path, e)



"""Contains all database models."""

from django.db import models
from django.db.models import QuerySet

import core.musiq.song_queue
import core.musiq.song_utils as song_utils


# Create your models here.
class Tag(models.Model):
    """Stores hashtags."""

    text = models.CharField(max_length=100)
    active = models.BooleanField()

    def __str__(self) -> str:
        return self.text


class Counter(models.Model):
    """Stores the visitors counter. Only has one elment."""

    value = models.IntegerField()

    def __str__(self) -> str:
        return str(self.value)


class ArchivedSong(models.Model):
    """Stores an archived song.
    url identifies the song uniquely in the database and on the internet (if applicable)."""

    url = models.CharField(max_length=2000, unique=True)
    artist = models.CharField(max_length=1000)
    title = models.CharField(max_length=1000)
    counter = models.IntegerField()

    def __str__(self) -> str:
        return self.title + " (" + self.url + "): " + str(self.counter)

    def displayname(self) -> str:
        """Formats the song using the utility method."""
        return song_utils.displayname(self.artist, self.title)


class ArchivedPlaylist(models.Model):
    """Stores an archived playlist.
    url identifies the playlist uniquely in the database and on the internet (if applicable)."""

    id: int
    entries: QuerySet
    list_id = models.CharField(max_length=2000)
    title = models.CharField(max_length=1000)
    created = models.DateTimeField(auto_now_add=True)
    counter = models.IntegerField()

    def __str__(self) -> str:
        return self.title + ": " + str(self.counter)


class PlaylistEntry(models.Model):
    """Stores an entry to a playlist. Connects ArchivedSong and ArchivedPlaylist."""

    playlist = models.ForeignKey(
        "ArchivedPlaylist", on_delete=models.CASCADE, related_name="entries"
    )
    index = models.IntegerField()
    url = models.CharField(max_length=2000)

    def __str__(self) -> str:
        return self.playlist.title + "[" + str(self.index) + "]: " + self.url

    class Meta:
        ordering = ["playlist", "index"]


class ArchivedQuery(models.Model):
    """Stores the queries from the musiq page and the ArchivedSong it lead to."""

    song = models.ForeignKey(
        "ArchivedSong", on_delete=models.CASCADE, related_name="queries"
    )
    query = models.CharField(max_length=1000)

    def __str__(self) -> str:
        return self.query


class ArchivedPlaylistQuery(models.Model):
    """Stores the queries from the musiq page and the ArchivedPlaylist it lead to."""

    playlist = models.ForeignKey(
        "ArchivedPlaylist", on_delete=models.CASCADE, related_name="queries"
    )
    query = models.CharField(max_length=1000)

    def __str__(self) -> str:
        return self.query


class QueuedSong(models.Model):
    """Stores a song in the song queue so the queue is not lost on server restart."""

    id: int
    index = models.IntegerField()
    manually_requested = models.BooleanField()
    votes = models.IntegerField(default=0)
    internal_url = models.CharField(max_length=2000, blank=True, null=True)
    external_url = models.CharField(max_length=2000)
    stream_url = models.CharField(max_length=2000, blank=True, null=True)
    artist = models.CharField(max_length=1000)
    title = models.CharField(max_length=1000)
    duration = models.IntegerField()
    objects = core.musiq.song_queue.SongQueue()

    def __str__(self) -> str:
        return str(self.index) + ": " + self.title + " (" + self.internal_url + ")"

    def displayname(self) -> str:
        """Formats the song using the utility method."""
        return song_utils.displayname(self.artist, self.title)

    class Meta:
        ordering = ["index"]


class CurrentSong(models.Model):
    """Stores the currently playing song. Only has one element."""

    queue_key = models.IntegerField()
    manually_requested = models.BooleanField()
    votes = models.IntegerField()
    artist = models.CharField(max_length=1000)
    title = models.CharField(max_length=1000)
    duration = models.IntegerField()
    internal_url = models.CharField(max_length=2000)
    external_url = models.CharField(max_length=2000)
    stream_url = models.CharField(max_length=2000, blank=True, null=True)
    created = models.DateTimeField(auto_now_add=True)
    last_paused = models.DateTimeField(auto_now_add=True)

    def __str__(self) -> str:
        return self.title + " (" + self.internal_url + ")"

    def displayname(self) -> str:
        """Formats the song using the utility method."""
        return song_utils.displayname(self.artist, self.title)


class RequestLog(models.Model):
    """Stores the request of a client and its result."""

    created = models.DateTimeField(auto_now_add=True)
    song = models.ForeignKey(
        "ArchivedSong", on_delete=models.SET_NULL, blank=True, null=True
    )
    playlist = models.ForeignKey(
        "ArchivedPlaylist", on_delete=models.SET_NULL, blank=True, null=True
    )
    address = models.CharField(max_length=50)

    def item_displayname(self) -> str:
        """Returns the displayname of the song or the title of the playlist"""
        if self.song is not None:
            return self.song.displayname()
        if self.playlist is not None:
            return self.playlist.title
        return "Unknown"

    def __str__(self) -> str:
        if self.song is not None:
            return self.address + ": " + self.song.displayname()
        if self.playlist is not None:
            return self.address + ": " + self.playlist.title
        return self.address + ": <None>"


class PlayLog(models.Model):
    """Stores the log of a played song."""

    created = models.DateTimeField(auto_now_add=True)
    song = models.ForeignKey(
        "ArchivedSong", on_delete=models.SET_NULL, blank=True, null=True
    )
    manually_requested = models.BooleanField()
    votes = models.IntegerField(null=True)

    def song_displayname(self) -> str:
        """Returns the displayname of the song (if present)"""
        if not self.song:
            return "Unknown"
        return self.song.displayname()

    def __str__(self) -> str:
        return (
            "played " + self.song_displayname() + " with " + str(self.votes) + " votes"
        )


class Setting(models.Model):
    """key value storage for persistent settings."""

    key = models.CharField(max_length=200, unique=True)
    value = models.CharField(max_length=200)

    def __str__(self) -> str:
        return self.key + ": " + ("None" if self.value is None else self.value)



"""This module handles the suggestions when starting to
type into the input field on the musiq page."""

from __future__ import annotations

import random
from typing import Dict, Union, List

from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponseBadRequest
from django.http.response import JsonResponse, HttpResponse
from watson import search as watson

import core.musiq.song_utils as song_utils
import core.settings.storage as storage
from core import redis
from core.models import ArchivedPlaylist, ArchivedSong
from core.musiq.song_provider import SongProvider


def random_suggestion(request: WSGIRequest) -> HttpResponse:
    """This method returns a random suggestion from the database.
    Depending on the value of :param playlist:,
    either a previously pushed playlist or song is returned."""
    suggest_playlist = request.GET["playlist"] == "true"
    if not suggest_playlist:
        if ArchivedSong.objects.count() == 0:
            return HttpResponseBadRequest("No songs to suggest from")
        index = random.randint(0, ArchivedSong.objects.count() - 1)
        song = ArchivedSong.objects.all()[index]
        return JsonResponse({"suggestion": song.displayname(), "key": song.id})

    # exclude radios from suggestions
    remaining_playlists = (
        ArchivedPlaylist.objects.all()
        .exclude(list_id__startswith="RD")
        .exclude(list_id__contains="&list=RD")
    )
    if remaining_playlists.count() == 0:
        return HttpResponseBadRequest("No playlists to suggest from")
    index = random.randint(0, remaining_playlists.count() - 1)
    playlist = remaining_playlists.all()[index]
    return JsonResponse({"suggestion": playlist.title, "key": playlist.id})


def _online_suggestions(query, suggest_playlist) -> List[Dict[str, Union[str, int]]]:
    results: List[Dict[str, Union[str, int]]] = []
    if storage.get("online_suggestions") and redis.get("has_internet"):
        if storage.get("spotify_enabled") and storage.get("spotify_suggestions") > 0:
            from core.musiq.spotify import Spotify

            spotify_suggestions = Spotify().get_search_suggestions(
                query, suggest_playlist
            )
            spotify_suggestions = spotify_suggestions[
                : storage.get("spotify_suggestions")
            ]
            for suggestion, external_url in spotify_suggestions:
                results.append(
                    {"key": external_url, "value": suggestion, "type": "spotify-online"}
                )

        if (
            storage.get("soundcloud_enabled")
            and storage.get("soundcloud_suggestions") > 0
        ):
            from core.musiq.soundcloud import Soundcloud

            soundcloud_suggestions = Soundcloud().get_search_suggestions(query)
            soundcloud_suggestions = soundcloud_suggestions[
                : storage.get("soundcloud_suggestions")
            ]
            for suggestion in soundcloud_suggestions:
                results.append(
                    {"key": -1, "value": suggestion, "type": "soundcloud-online"}
                )

        if storage.get("jamendo_enabled") and storage.get("jamendo_suggestions") > 0:
            from core.musiq.jamendo import Jamendo

            jamendo_suggestions = Jamendo().get_search_suggestions(query)
            jamendo_suggestions = jamendo_suggestions[
                : storage.get("jamendo_suggestions")
            ]
            for suggestion in jamendo_suggestions:
                results.append(
                    {"key": -1, "value": suggestion, "type": "jamendo-online"}
                )

        if storage.get("youtube_enabled") and storage.get("youtube_suggestions") > 0:
            from core.musiq.youtube import Youtube

            youtube_suggestions = Youtube().get_search_suggestions(query)
            youtube_suggestions = youtube_suggestions[
                : storage.get("youtube_suggestions")
            ]
            for suggestion in youtube_suggestions:
                results.append(
                    {"key": -1, "value": suggestion, "type": "youtube-online"}
                )
    return results


def get_suggestions(request: WSGIRequest) -> JsonResponse:
    """Returns suggestions for a given query.
    Combines online and offline suggestions."""
    query = request.GET["term"]
    suggest_playlist = request.GET["playlist"] == "true"

    if storage.get("new_music_only") and not suggest_playlist:
        return JsonResponse([], safe=False)

    results = _online_suggestions(query, suggest_playlist)

    if suggest_playlist:
        search_results = watson.search(query, models=(ArchivedPlaylist,))[
            : storage.get("number_of_suggestions")
        ]

        for playlist in search_results:
            playlist_info = playlist.meta
            archived_playlist = ArchivedPlaylist.objects.get(id=playlist_info["id"])
            result_dict: Dict[str, Union[str, int]] = {
                "key": playlist_info["id"],
                "value": playlist_info["title"],
                "counter": playlist.object.counter,
                "type": song_utils.determine_playlist_type(archived_playlist),
            }
            results.append(result_dict)
    else:
        search_results = watson.search(query, models=(ArchivedSong,))[
            : storage.get("number_of_suggestions")
        ]

        for search_result in search_results:
            song_info = search_result.meta

            if song_utils.is_forbidden(song_info["artist"]) or song_utils.is_forbidden(
                song_info["title"]
            ):
                continue

            try:
                provider = SongProvider.create(external_url=song_info["url"])
            except NotImplementedError:
                # For this song a provider is necessary that is not available
                # e.g. the song was played before, but the provider was disabled
                continue
            cached = provider.check_cached()
            # don't suggest online songs when we don't have internet
            if not redis.get("has_internet") and not cached:
                continue
            if provider.type == "local":
                # don't suggest local songs if they are not cached (=not at expected location)
                if not cached:
                    continue
            else:
                # don't suggest songs if the respective platform is disabled
                if not storage.get(f"{provider.type}_enabled"):
                    continue
            result_dict = {
                "key": song_info["id"],
                "value": song_utils.displayname(
                    song_info["artist"], song_info["title"]
                ),
                "counter": search_result.object.counter,
                "type": provider.type,
            }
            # Add duration for songs where it is available (=cached songs)
            if cached:
                metadata = provider.get_metadata()
                result_dict["durationFormatted"] = song_utils.format_seconds(
                    metadata["duration"]
                )
            results.append(result_dict)
        # mark suggestions whose displayname is identical
        seen_values = {}
        for i, result in enumerate(results):
            if result["value"] in seen_values:
                result["confusable"] = True
                results[seen_values[result["value"]]]["confusable"] = True
            seen_values[result["value"]] = i

    return JsonResponse(results, safe=False)



"""This module contains all public api endpoints."""
import re
from django.conf import settings as conf
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse, HttpResponseBadRequest
from django.views.decorators.csrf import csrf_exempt

import core.settings.storage as storage
from core import user_manager
from core.musiq import musiq


@csrf_exempt
def post_song(request: WSGIRequest) -> HttpResponse:
    """This endpoint is part of the API and exempt from CSRF checks.
    Shareberry uses this endpoint."""
    # only get ip on user requests
    request_ip = user_manager.get_client_ip(request)
    query = request.POST.get("query")
    if not query:
        return HttpResponseBadRequest("No query to share.")

    match = re.search(r"(?P<url>https?://[^\s]+)", query)
    if match:
        query = match.group("url")

    # Set the requested platform to 'spotify'.
    # It will automatically fall back to Youtube
    # if Spotify is not enabled or a youtube link was requested.
    successful, message, _ = musiq.do_request_music(
        request_ip, query, None, False, "spotify"
    )
    if not successful:
        return HttpResponseBadRequest(message)
    return HttpResponse(message)


def version(request: WSGIRequest) -> HttpResponse:
    """Return the version of the running instance."""
    return HttpResponse(f"Raveberry version {conf.VERSION}")



"""This module handles all basic settings."""

from __future__ import annotations

import subprocess

from django.core.handlers.wsgi import WSGIRequest

from core import user_manager, redis
from core.settings import storage
from core.settings.settings import control
from core.musiq import playback


def start() -> None:
    """Initializes this module. Checks whether internet is accessible."""
    _check_internet()


def _check_internet() -> None:
    response = subprocess.call(
        ["ping", "-c", "1", "-W", "3", "1.1.1.1"], stdout=subprocess.DEVNULL
    )
    if response == 0:
        redis.set("has_internet", True)
    else:
        redis.set("has_internet", False)


@control
def set_voting_enabled(request: WSGIRequest) -> None:
    """Enables or disables voting based on the given value."""
    enabled = request.POST.get("value") == "true"
    storage.set("voting_enabled", enabled)


@control
def set_new_music_only(request: WSGIRequest) -> None:
    """Enables or disables the new music only mode based on the given value."""
    enabled = request.POST.get("value") == "true"
    storage.set("new_music_only", enabled)


@control
def set_logging_enabled(request: WSGIRequest) -> None:
    """Enables or disables logging of requests and play logs based on the given value."""
    enabled = request.POST.get("value") == "true"
    storage.set("logging_enabled", enabled)


@control
def set_hashtags_active(request: WSGIRequest) -> None:
    """Enables or disables logging of requests and play logs based on the given value."""
    enabled = request.POST.get("value") == "true"
    storage.set("hashtags_active", enabled)


@control
def set_embed_stream(request: WSGIRequest) -> None:
    """Enables or disables logging of requests and play logs based on the given value."""
    enabled = request.POST.get("value") == "true"
    storage.set("embed_stream", enabled)


@control
def set_dynamic_embedded_stream(request: WSGIRequest) -> None:
    """Enables or disables logging of requests and play logs based on the given value."""
    enabled = request.POST.get("value") == "true"
    storage.set("dynamic_embedded_stream", enabled)


@control
def set_online_suggestions(request: WSGIRequest) -> None:
    """Enables or disables online suggestions based on the given value."""
    enabled = request.POST.get("value") == "true"
    storage.set("online_suggestions", enabled)


@control
def set_number_of_suggestions(request: WSGIRequest) -> None:
    """Set the number of archived suggestions based on the given value."""
    value = int(request.POST.get("value"))  # type: ignore
    storage.set("number_of_suggestions", value)


@control
def set_people_to_party(request: WSGIRequest) -> None:
    """Sets the amount of active clients needed to enable partymode."""
    value = int(request.POST.get("value"))  # type: ignore
    storage.set("people_to_party", value)


@control
def set_alarm_probability(request: WSGIRequest) -> None:
    """Sets the probability with which an alarm is triggered after each song."""
    value = float(request.POST.get("value"))  # type: ignore
    storage.set("alarm_probability", value)


@control
def set_buzzer_cooldown(request: WSGIRequest) -> None:
    """Sets the minimum time that needs to pass between buzzer presses."""
    value = int(request.POST.get("value"))  # type: ignore
    storage.set("buzzer_cooldown", value)


@control
def trigger_alarm(_request: WSGIRequest) -> None:
    """Manually triggers an alarm."""
    playback.trigger_alarm()
    # because a state update is sent after every control (including this one)
    # a state update with alarm not being set would be sent
    # prevent this by manually setting this redis variable prematurely
    redis.set("alarm_playing", True)


@control
def set_downvotes_to_kick(request: WSGIRequest) -> None:
    """Sets the number of downvotes that are needed to remove a song from the queue."""
    value = int(request.POST.get("value"))  # type: ignore
    storage.set("downvotes_to_kick", value)


@control
def set_max_download_size(request: WSGIRequest) -> None:
    """Sets the maximum amount of MB that are allowed for a song that needs to be downloaded."""
    value = float(request.POST.get("value"))  # type: ignore
    storage.set("max_download_size", value)


@control
def set_max_playlist_items(request: WSGIRequest) -> None:
    """Sets the maximum number of songs that are downloaded from a playlist."""
    value = int(request.POST.get("value"))  # type: ignore
    storage.set("max_playlist_items", value)


@control
def set_max_queue_length(request: WSGIRequest) -> None:
    """Sets the maximum number of songs that are downloaded from a playlist."""
    value = int(request.POST.get("value"))  # type: ignore
    storage.set("max_queue_length", value)


@control
def set_additional_keywords(request: WSGIRequest):
    """Sets the keywords to filter out of results."""
    value = request.POST.get("value")
    storage.set("additional_keywords", value)


@control
def set_forbidden_keywords(request: WSGIRequest):
    """Sets the keywords to filter out of results."""
    value = request.POST.get("value")
    storage.set("forbidden_keywords", value)


@control
def check_internet(_request: WSGIRequest) -> None:
    """Checks whether an internet connection exists and updates the internal state."""
    _check_internet()


@control
def update_user_count(_request: WSGIRequest) -> None:
    """Force an update on the active user count."""
    user_manager.update_user_count()



"""This module handles all settings related to system configuration."""
from __future__ import annotations

import configparser
import os
import re
import shutil
import subprocess
from typing import Dict, Optional, Tuple

import cachetools.func
import redis.exceptions
import requests
from django.conf import settings
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse, HttpResponseBadRequest

from core.musiq.playback import player_lock
from core.settings import storage
from core.settings.settings import control


def _restart_mopidy() -> None:
    subprocess.call(["sudo", "/usr/local/sbin/raveberry/restart_mopidy"])
    try:
        player_lock.release()
    except redis.exceptions.LockError:
        # the lock was already released
        pass


def update_mopidy_config(output: str) -> None:
    """Updates mopidy's config with the credentials stored in the database.
    If no config_file is given, the default one is used."""
    if settings.DOCKER:
        # raveberry cannot restart mopidy in the docker setup
        return

    if output == "pulse":
        if storage.get("feed_cava") and shutil.which("cava"):
            output = "cava"
        else:
            output = "regular"

    spotify_username = storage.get("spotify_username")
    spotify_password = storage.get("spotify_password")
    spotify_client_id = storage.get("spotify_client_id")
    spotify_client_secret = storage.get("spotify_client_secret")
    soundcloud_auth_token = storage.get("soundcloud_auth_token")
    jamendo_client_id = storage.get("jamendo_client_id")

    subprocess.call(
        [
            "sudo",
            "/usr/local/sbin/raveberry/update_mopidy_config",
            output,
            spotify_username,
            spotify_password,
            spotify_client_id,
            spotify_client_secret,
            soundcloud_auth_token,
            jamendo_client_id,
        ]
    )
    _restart_mopidy()


def check_mopidy_extensions() -> Dict[str, Tuple[bool, str]]:
    """Returns a dict indicating for each extension whether it is enabled
    and provides a message.
    Handles both service and user mopidy instances."""
    try:
        subprocess.check_call(
            ["systemctl", "is-active", "mopidy"], stdout=subprocess.DEVNULL
        )
    except subprocess.CalledProcessError:
        extensions = _check_mopidy_extensions_user()
    else:
        extensions = _check_mopidy_extensions_service()
    return extensions


def _check_mopidy_extensions_user() -> Dict[str, Tuple[bool, str]]:
    config = subprocess.run(
        ["mopidy", "config"],
        stdout=subprocess.PIPE,
        universal_newlines=True,
        check=True,
    ).stdout
    parser = configparser.ConfigParser()
    parser.read_string(config)
    extensions = {}
    for extension in ["spotify", "soundcloud", "jamendo"]:
        try:
            if parser[extension]["enabled"] == "true":
                extensions[extension] = (True, "Extension probably functional")
            else:
                extensions[extension] = (False, "Extension disabled")
        except KeyError:
            extensions[extension] = (False, "Extension disabled")
    return extensions


def _check_mopidy_extensions_service() -> Dict[str, Tuple[bool, str]]:
    # check the mopidy log and see if there is a spotify login error since the last restart
    log = subprocess.check_output(
        ["sudo", "/usr/local/sbin/raveberry/read_mopidy_log"], universal_newlines=True
    )

    extensions = {}
    for line in log.split("\n")[::-1]:
        if "spotify" not in extensions:

            if (
                line.startswith("ERROR")
                and "spotify.session" in line
                and "USER_NEEDS_PREMIUM"
            ):
                extensions["spotify"] = (False, "Spotify Premium is required")
            elif line.startswith("ERROR") and "spotify.session" in line:
                extensions["spotify"] = (False, "User or Password are wrong")
            elif line.startswith("ERROR") and "mopidy_spotify.web" in line:
                extensions["spotify"] = (
                    False,
                    "Client ID or Client Secret are wrong or expired",
                )
            elif (
                line.startswith("WARNING")
                and "spotify" in line
                and "The extension has been automatically disabled" in line
            ):
                extensions["spotify"] = (False, "Configuration Error")

        if "soundcloud" not in extensions:
            if line.startswith("ERROR") and 'Invalid "auth_token"' in line:
                extensions["soundcloud"] = (False, "auth_token is invalid")
            elif (
                line.startswith("WARNING")
                and "soundcloud" in line
                and "The extension has been automatically disabled" in line
            ):
                extensions["soundcloud"] = (False, "Configuration Error")

        if "jamendo" not in extensions:
            if line.startswith("ERROR") and 'Invalid "client_id"' in line:
                extensions["jamendo"] = (False, "client_id is invalid")
            elif (
                line.startswith("WARNING")
                and "jamendo" in line
                and "The extension has been automatically disabled" in line
            ):
                extensions["jamendo"] = (False, "Configuration Error")

        if (
            "spotify" in extensions
            and "soundcloud" in extensions
            and "jamendo" in extensions
        ):
            break

        if line.startswith("Started Mopidy music server."):
            if "spotify" not in extensions:
                extensions["spotify"] = (True, "Login successful")
            if "soundcloud" not in extensions:
                extensions["soundcloud"] = (True, "auth_token valid")
            if "jamendo" not in extensions:
                extensions["jamendo"] = (True, "client_id could not be checked")
            break
    else:
        # there were too many lines in the log, could not determine whether there was an error
        if "spotify" not in extensions:
            extensions["spotify"] = (True, "No info found, enabling to be safe")
        if "soundcloud" not in extensions:
            extensions["soundcloud"] = (True, "No info found, enabling to be safe")
        if "jamendo" not in extensions:
            extensions["jamendo"] = (True, "No info found, enabling to be safe")
    return extensions


@control
def disable_hotspot(_request: WSGIRequest) -> None:
    """Disable the wifi created by Raveberry."""
    subprocess.call(["sudo", "/usr/local/sbin/raveberry/disable_hotspot"])


@control
def enable_hotspot(_request: WSGIRequest) -> None:
    """Enable the wifi created by Raveberry."""
    subprocess.call(["sudo", "/usr/local/sbin/raveberry/enable_hotspot"])


@control
def disable_wifi_protection(_request: WSGIRequest) -> None:
    """Disable password protection of the hotspot, making it public."""
    subprocess.call(["sudo", "/usr/local/sbin/raveberry/disable_wifi_protection"])


@control
def enable_wifi_protection(_request: WSGIRequest) -> None:
    """Enable password protection of the hotspot.
    The password was defined during setup."""
    subprocess.call(["sudo", "/usr/local/sbin/raveberry/enable_wifi_protection"])


@control
def disable_tunneling(_request: WSGIRequest) -> None:
    """Disable forwarding of packets to the other network (probably the internet)."""
    subprocess.call(["sudo", "/usr/local/sbin/raveberry/disable_tunneling"])


@control
def enable_tunneling(_request: WSGIRequest) -> None:
    """Enable forwarding of packets to the other network.
    Enables clients connected to the hotspot to browse the internet (if available)."""
    subprocess.call(["sudo", "/usr/local/sbin/raveberry/disable_tunneling"])
    subprocess.call(["sudo", "/usr/local/sbin/raveberry/enable_tunneling"])


@control
def disable_remote(_request: WSGIRequest) -> None:
    """Disables the connection to an external server."""
    subprocess.call(["sudo", "/usr/local/sbin/raveberry/disable_remote"])


@control
def enable_remote(_request: WSGIRequest) -> None:
    """Enables the connection to an external server.
    Allows this instance to be reachable from a public domain."""
    subprocess.call(["sudo", "/usr/local/sbin/raveberry/enable_remote"])


@control
def restart_server(_request: WSGIRequest) -> None:
    """Restarts the server."""
    subprocess.call(["sudo", "/usr/local/sbin/raveberry/restart_server"])


@control
def kill_workers(_request: WSGIRequest) -> None:
    """Force kills all celery workers."""
    subprocess.call(["sudo", "/usr/local/sbin/raveberry/kill_workers"])


@control
def reboot_system(_request: WSGIRequest) -> None:
    """Reboots the system."""
    subprocess.call(["sudo", "/usr/local/sbin/raveberry/reboot_system"])


@control
def shutdown_system(_request: WSGIRequest) -> None:
    """Shuts down the system."""
    subprocess.call(["sudo", "/usr/local/sbin/raveberry/shutdown_system"])


@cachetools.func.ttl_cache(ttl=60 * 60 * 24)
def fetch_latest_version() -> Optional[str]:
    """Looks up the newest version number from PyPi and returns it."""
    # https://github.com/pypa/pip/issues/9139
    p = subprocess.run(
        "pip3 install --use-deprecated=legacy-resolver raveberry==nonexistingversion".split(),
        stdout=subprocess.DEVNULL,
        stderr=subprocess.PIPE,
        universal_newlines=True,
    )
    if p.returncode == 2:
        # pip does not now the --use-deprecated=legacy-resolver
        # probably an older version that does not need it
        p = subprocess.run(
            "pip3 install raveberry==nonexistingversion".split(),
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )

    # parse the newest verson from pip output
    for line in p.stderr.splitlines():
        if "from versions" in line:
            versions = [re.sub(r"[^0-9.]", "", token) for token in line.split()]
            versions = [version for version in versions if version]
            latest_version = versions[-1]
            return latest_version
    else:
        return None


@control
def get_latest_version(_request: WSGIRequest) -> HttpResponse:
    """Returns the newest version number of Raveberry from PyPi."""
    latest_version = fetch_latest_version()
    if latest_version is None:
        return HttpResponseBadRequest("Could not determine latest version.")
    return HttpResponse(latest_version)


@control
def get_changelog(_request: WSGIRequest) -> HttpResponse:
    """Retreives the changelog and returns it."""
    changelog = requests.get(
        "https://raw.githubusercontent.com/raveberry/raveberry/master/CHANGELOG.md"
    ).text
    return HttpResponse(changelog)


@control
def get_upgrade_config(_request: WSGIRequest) -> HttpResponse:
    """Returns the config that will be used for the upgrade."""
    with open(os.path.join(settings.BASE_DIR, "config/raveberry.yaml")) as f:
        config = f.read()
    lines = config.splitlines()
    lines = [line for line in lines if not line.startswith("#")]
    return HttpResponse("\n".join(lines))


@control
def upgrade_raveberry(_request: WSGIRequest) -> HttpResponse:
    """Performs an upgrade of raveberry."""

    subprocess.call(["sudo", "/usr/local/sbin/raveberry/start_upgrade_service"])

    return HttpResponse("Upgrading... Look for logs in /var/www/")



"""This module contains all Jamendo related code."""

from __future__ import annotations

import logging
from contextlib import closing
from typing import Optional, List, TYPE_CHECKING
from urllib.parse import urlparse

import requests
from django.http.response import HttpResponse

import core.settings.storage as storage
from core.musiq import song_utils
from core.musiq import musiq
from core.musiq.song_provider import SongProvider
from core.musiq.playlist_provider import PlaylistProvider

if TYPE_CHECKING:
    from core.musiq.song_utils import Metadata


class JamendoClient:
    """Interface with the Jamendo API. Taken from mopidy-jamendo."""

    def __init__(self, client_id: str) -> None:
        self.session = requests.Session()
        self.session.headers.update(
            {"user-agent": "Mopidy-Jamendo/0.1.0 Mopidy/3.0.2 CPython/3.7.5"}
        )
        self.client_id = client_id

    def get(self, url: str, params: dict = None) -> dict:
        """Perform the specified API request."""
        url = f"https://api.jamendo.com/v3.0/{url}"
        if not params:
            params = {}
        params["client_id"] = self.client_id
        try:
            with closing(self.session.get(url, params=params)) as res:
                logging.debug(f"Requested {res.url}")
                res.raise_for_status()
                return res.json()
        except Exception as e:
            if isinstance(e, requests.HTTPError) and e.response.status_code == 401:
                logging.error('Invalid "client_id" used for Jamendo authentication!')
            else:
                logging.error(f"Jamendo API request failed: {e}")
        return {}


class Jamendo:
    """This class contains code for both the song and playlist provider"""

    _web_client: JamendoClient = None  # type: ignore

    @staticmethod
    def _get_web_client() -> JamendoClient:
        if Jamendo._web_client is None:
            client_id = storage.get(key="jamendo_client_id")
            Jamendo._web_client = JamendoClient(client_id=client_id)
        return Jamendo._web_client

    @property
    def web_client(self) -> JamendoClient:
        """Returns the web client if it was already created.
        If not, it is created using the client-id from mopidy-jamendo."""
        return Jamendo._get_web_client()

    def get_search_suggestions(self, query: str) -> List[str]:
        """Returns a list of suggested items for the given query."""

        if len(query.strip()) <= 1:
            # query length needs to be at least two characters
            return []

        result = self.web_client.get(
            "autocomplete", params={"prefix": query, "limit": "20"}
        )

        try:
            suggestions = result["results"]["tracks"]
        except (KeyError, TypeError):
            return []

        suggestions = [
            suggestion
            for suggestion in suggestions
            if suggestion != query and not song_utils.is_forbidden(suggestion)
        ]
        return suggestions


class JamendoSongProvider(SongProvider, Jamendo):
    """This class handles songs from Soundcloud."""

    @staticmethod
    def get_id_from_external_url(url: str) -> str:
        return [
            component
            for component in urlparse(url).path.split("/")
            if component.isdigit()
        ][0]

    @staticmethod
    def get_id_from_internal_url(url: str) -> str:
        """Returns the internal id based on the given url."""
        return url.split(":")[-1]

    def __init__(self, query: Optional[str], key: Optional[int]) -> None:
        self.type = "jamendo"
        super().__init__(query, key)

        self.metadata: "Metadata" = {}
        self.external_url = None

    def check_cached(self) -> bool:
        # Jamendo songs cannot be cached and have to be streamed everytime
        return False

    def check_available(self) -> bool:
        if not self.gather_metadata():
            return False
        # the default bitrate in jamendo is 96kbs
        # estimate the size of a song by multiplying with its duration
        size = self.metadata["duration"] * 96 / 8 * 1000
        return self.check_not_too_large(size)

    def gather_metadata(self) -> bool:
        """Fetches metadata for this song's uri from Jamendo."""
        if not self.id:
            results = self.web_client.get("tracks", {"search": self.query})["results"]

            # apply the filterlist from the settings
            for item in results:
                artist = item["artist_name"]
                title = item["name"]
                if song_utils.is_forbidden(artist) or song_utils.is_forbidden(title):
                    continue
                result = item
                break
            else:
                # all tracks got filtered
                return False
            self.id = result["id"]
        else:
            try:
                result = self.web_client.get("tracks", {"id": self.id})["results"][0]
            except (KeyError, IndexError):
                self.error = f"id {self.id} not found"
                return False
        self.metadata["internal_url"] = self.get_internal_url()
        self.metadata["external_url"] = result["shareurl"]
        self.metadata["stream_url"] = result["audio"]
        self.metadata["artist"] = result["artist_name"]
        self.metadata["title"] = result["name"]
        self.metadata["duration"] = result["duration"]
        return True

    def get_metadata(self) -> "Metadata":
        if not self.metadata:
            self.gather_metadata()
        return self.metadata

    def _get_path(self) -> str:
        # Jamendo is not cached in the cache directory
        raise NotImplementedError()

    def get_internal_url(self) -> str:
        if not self.id:
            raise ValueError()
        return f"jamendo:track:{self.id}"

    def get_external_url(self) -> str:
        if not self.id:
            raise ValueError()
        return "https://www.jamendo.com/track/" + self.id

    def get_suggestion(self) -> str:
        result = self.web_client.get(
            "tracks/similar", params={"id": self.id, "limit": "1"}
        )

        try:
            external_url = result["results"][0]["shareurl"]
        except IndexError:
            self.error = "no recommendation found"
            raise ValueError("No suggested track")

        return external_url

    def request_radio(self, request_ip: str) -> HttpResponse:

        result = self.web_client.get(
            "recommendations",
            params={"id": self.id, "limit": storage.get("basic.max_playlist_items")},
        )

        for track in result["results"]:
            external_url = track["shareurl"]
            musiq.do_request_music(
                "",
                external_url,
                None,
                False,
                "jamendo",
                archive=False,
                manually_requested=False,
            )

        return HttpResponse("queueing radio")


class JamendoPlaylistProvider(PlaylistProvider, Jamendo):
    """This class handles Jamendo Playlists."""

    @staticmethod
    def get_id_from_external_url(url: str) -> Optional[str]:
        if not url.startswith("https://www.jamendo.com/"):
            return None
        return [
            component
            for component in urlparse(url).path.split("/")
            if component.isdigit()
        ][0]

    def __init__(self, query: Optional[str], key: Optional[int]) -> None:
        self.type = "jamendo"
        super().__init__(query, key)

    def search_id(self) -> Optional[str]:
        results = self.web_client.get(
            "playlists", params={"namesearch": self.query, "limit": 1}
        )

        try:
            playlist = results["results"][0]
        except IndexError:
            self.error = "No playlist found"
            return None

        list_id = playlist["id"]
        self.title = playlist["name"]

        return list_id

    def is_radio(self) -> bool:
        return False

    def fetch_metadata(self) -> bool:
        if self.title is None:
            result = self.web_client.get("playlists", params={"id": self.id})
            self.title = result["results"][0]["name"]

        results = self.web_client.get("playlists/tracks", params={"id": self.id})
        tracks = results["results"][0]["tracks"]

        for track in tracks:
            url = "https://www.jamendo.com/track/" + track["id"]
            self.urls.append(url)

        return True



"""This module contains the celery app."""
import os
from celery import Celery

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "main.settings")

app = Celery("core")

app.config_from_object("django.conf:settings")


class CeleryNotReachable(Exception):
    """Raised when celery should be reachable but is not."""

    pass


def start() -> None:
    """Initializes celery."""
    # check if celery is up and wait for a maximum of 5 seconds
    for _ in range(10):
        if app.control.ping(timeout=0.5):
            break
    else:
        raise CeleryNotReachable("Celery worker pool not reachable. Is it running?")

    # stop running celery tasks from old django instance
    active_tasks = app.control.inspect().active()
    if active_tasks:
        for hostname, tasks in active_tasks.items():
            for task in tasks:
                app.control.revoke(task_id=task["id"], terminate=True)



"""This module handles requests for the network info page."""

from __future__ import annotations

import io
import subprocess
from typing import Any, Dict

import qrcode
from django.contrib.auth.decorators import user_passes_test
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse
from django.shortcuts import render

from core import util, base, user_manager


def state_dict() -> Dict[str, Any]:
    state = base.state_dict()
    return state


def _qr_path(data) -> str:
    # from https://github.com/lincolnloop/python-qrcode/blob/master/qrcode/console_scripts.py
    module = "qrcode.image.svg.SvgPathImage"
    module, name = module.rsplit(".", 1)
    imp = __import__(module, {}, [], [name])
    SvgPathImage = getattr(imp, name)

    qr = qrcode.QRCode()
    qr.add_data(data)
    img = qr.make_image(image_factory=SvgPathImage)
    with io.BytesIO() as stream:
        img.save(stream)
        svg = stream.getvalue().decode()
    tag = svg.split("\n")[1]
    return tag


@user_passes_test(user_manager.has_controls)
def index(request: WSGIRequest) -> HttpResponse:
    """Renders the /network_info page. Only admin is allowed to see this page."""
    context = base.context(request)

    # hotspot information
    context["hotspot_enabled"] = False
    try:
        if subprocess.call(["/usr/local/sbin/raveberry/hotspot_enabled"]) != 0:
            # Hotspot is enabled, show its info as well
            context["hotspot_enabled"] = True

            with open("/etc/hostapd/hostapd_protected.conf") as f:
                for line in f:
                    line = line.strip()
                    if line.startswith("ssid"):
                        ssid = line.split("=")[1]
                    if line.startswith("wpa_passphrase"):
                        password = line.split("=")[1]

            device = util.get_devices()[-1]
            ip = util.ip_of_device(device)
            url = f"http://{ip}/"

            context["hotspot_ssid"] = ssid
            context["hotspot_password"] = password
            context["hotspot_wifi_qr"] = _qr_path(f"WIFI:S:{ssid};T:WPA;P:{password};;")
            context["hotspot_url"] = url
            context["hotspot_url_qr"] = _qr_path(url)
            context["hotspot_ip"] = ip
    except FileNotFoundError:
        # hotspot was not configured
        pass

    # connected network information
    ssid = None
    password = None
    try:
        ssid = subprocess.check_output(
            "/sbin/iwgetid --raw".split(), universal_newlines=True
        )[:-1]
        wifi_active = True
    except (subprocess.CalledProcessError, FileNotFoundError):
        wifi_active = False

    if wifi_active:
        try:
            password = subprocess.check_output(
                ["sudo", "/usr/local/sbin/raveberry/password_for_ssid", ssid],
                universal_newlines=True,
            )
        except subprocess.CalledProcessError:
            pass

    device = util.get_devices()[0]
    ip = util.ip_of_device(device)

    wifi_qr = _qr_path(f"WIFI:S:{ssid};T:WPA;P:{password};;")
    raveberry_url = f"http://{ip}/"
    raveberry_qr = _qr_path(raveberry_url)

    if wifi_active:
        context["ssid"] = ssid
        context["password"] = password
        context["wifi_qr"] = wifi_qr
    else:
        context["ssid"] = None
        context["password"] = None
        context["wifi_qr"] = None
    context["url"] = raveberry_url
    context["url_qr"] = raveberry_qr
    context["ip"] = ip
    return render(request, "network_info.html", context)



# Generated by Django 2.2.20 on 2021-07-02 19:50

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [("core", "0011_auto_20210427_1339")]

    operations = [
        migrations.AddField(
            model_name="currentsong",
            name="last_paused",
            field=models.DateTimeField(
                auto_now_add=True, default=django.utils.timezone.now
            ),
            preserve_default=False,
        )
    ]
